"""Contains plot tse action handling."""

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.channels import get_channels_by_type

from meggie.mainwindow.dynamic import Action

from meggie.utilities.dialogs.TFROutputOptionsMain import TFROutputOptions

from meggie.actions.tfr_plot_tse.controller.tfr import plot_tse_averages
from meggie.actions.tfr_plot_tse.controller.tfr import plot_tse_topo


class PlotTSE(Action):

    def run(self, params={}):
        """Plots a TSE from TFR object."""
        try:
            selected_name = self.data["outputs"]["tfr"][0]
        except IndexError:
            return

        def option_handler(params):
            params["channel_groups"] = self.experiment.channel_groups
            params["name"] = selected_name

            try:
                self.handler(self.experiment.active_subject, params)
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = TFROutputOptions(
            self.window,
            self.experiment,
            selected_name,
            handler=option_handler,
            ask_condition=False,
        )
        dialog.show()

    def handler(self, subject, params):
        """ """
        info = subject.tfr[params["name"]].info
        if params["output_option"] == "all_channels":
            chs = list(get_channels_by_type(info).keys())
            if "eeg" in chs:
                plot_tse_topo(
                    subject,
                    params["name"],
                    params["blmode"],
                    params["blstart"],
                    params["blend"],
                    params["tmin"],
                    params["tmax"],
                    params["fmin"],
                    params["fmax"],
                    "eeg",
                )
            if "grad" in chs or "mag" in chs:
                plot_tse_topo(
                    subject,
                    params["name"],
                    params["blmode"],
                    params["blstart"],
                    params["blend"],
                    params["tmin"],
                    params["tmax"],
                    params["fmin"],
                    params["fmax"],
                    "meg",
                )
        else:
            plot_tse_averages(
                subject,
                params["name"],
                params["blmode"],
                params["blstart"],
                params["blend"],
                params["tmin"],
                params["tmax"],
                params["fmin"],
                params["fmax"],
                params["channel_groups"],
            )



"""Contains implementation for raw plot"""

import logging

import matplotlib.pyplot as plt
import numpy as np

import mne

from meggie.utilities.events import find_stim_channel
from meggie.utilities.events import find_events
from meggie.utilities.messaging import exc_messagebox

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class PlotRaw(Action):
    """Shows a mne raw plot and saves the changes to FS
    if annotations or bads have changed.
    """

    def run(self, params={}):

        subject = self.experiment.active_subject
        raw = subject.get_raw()
        if not raw:
            return

        # Create a copy to not update the original yet
        raw = raw.copy()

        old_bads = raw.info["bads"].copy()
        old_annotations = raw.annotations.copy()

        # find events
        stim_ch = find_stim_channel(raw)
        if not stim_ch:
            events = None
        else:
            events = find_events(raw, stim_ch=stim_ch)

        def handle_close(event):
            bads_changed = sorted(raw.info["bads"]) != sorted(old_bads)

            annotations_changed = False
            if len(raw.annotations) != len(old_annotations):
                annotations_changed = True
            elif not np.allclose(raw.annotations.onset, old_annotations.onset):
                annotations_changed = True

            params = {}
            params["bads_changed"] = bads_changed
            params["bads"] = [str(bad) for bad in raw.info["bads"]]

            params["annotations_changed"] = annotations_changed

            params["annotations"] = []
            for idx in range(len(raw.annotations)):
                params["annotations"].append(
                    {
                        "onset": raw.annotations.onset[idx],
                        "duration": raw.annotations.duration[idx],
                        "description": raw.annotations.description[idx],
                    }
                )

            try:
                # only run logged action if there are changes
                if bads_changed or annotations_changed:
                    self.handler(subject, params)
                    self.window.initialize_ui()
            except Exception as exc:
                exc_messagebox(self.window, exc)

        fig = raw.plot(events=events, show=False)
        fig.canvas.mpl_connect("close_event", handle_close)
        plt.show()

    @subject_action
    def handler(self, subject, params):

        if params["bads_changed"]:
            logging.getLogger("ui_logger").info("Bads changed!")
        if params["annotations_changed"]:
            logging.getLogger("ui_logger").info("Annotations changed!")

        raw = subject.get_raw()

        # Finally, update the raw based on new params. First annotations.
        onset = []
        duration = []
        description = []
        for annotation_params in params["annotations"]:
            # set_annotations adds the first_samp based offset, so take that into account.
            onset.append(
                annotation_params["onset"] - raw.first_samp / raw.info["sfreq"]
            )
            duration.append(annotation_params["duration"])
            description.append(annotation_params["description"])
        annotations = mne.Annotations(onset, duration, description)
        raw.set_annotations(annotations)

        # Then the bad channels.
        raw.info["bads"] = params["bads"]

        subject.save()



import tempfile
import os
import numpy as np
import mne

from meggie.utilities.filemanager import save_csv
from meggie.utilities.filemanager import load_csv
from meggie.utilities.filemanager import open_raw
from meggie.utilities.filemanager import save_raw
from meggie.utilities.filemanager import get_supported_formats


def test_save_and_load_csv():
    with tempfile.TemporaryDirectory() as dirpath:

        filepath = os.path.join(dirpath, "data.csv")

        data = np.array([[1, 2, 3], [4, 5, 6]])
        column_names = ["A", "B", "C"]
        row_descs = [("X", "Kissa"), ("Y", "Koira")]

        save_csv(filepath, data, column_names, row_descs)

        loaded_column_names, loaded_row_descs, loaded_data = load_csv(filepath)

        assert loaded_column_names == column_names
        assert loaded_row_descs[0] == row_descs[0]
        assert loaded_row_descs[1] == row_descs[1]
        assert np.array_equal(loaded_data, data)


def test_save_and_load_raw():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, "MEG", "sample", "sample_audvis_raw.fif")
    sample_raw = mne.io.read_raw_fif(sample_fname, preload=True)
    raw_copy = sample_raw.copy()

    # make long enough raw to test split files
    raw = mne.concatenate_raws([sample_raw] + 8 * [raw_copy])

    with tempfile.TemporaryDirectory() as dirpath:
        path = os.path.join(dirpath, "raw.fif")
        save_raw(raw, path)

        # release memory before opening again
        sample_raw = None
        raw_copy = None
        raw = None

        open_raw(path)


def test_supported_formats():
    supported_formats = get_supported_formats()
    assert isinstance(supported_formats, list)
    assert isinstance(supported_formats[0], tuple)
    assert isinstance(supported_formats[0][0], str)
    assert isinstance(supported_formats[0][1], list)



"""Contains functions for tasks related to file system."""

import appdirs
import os
import shutil
import re
import logging

import importlib.resources as _resources
import importlib.metadata as _metadata

import numpy as np
import mne

from mne.io._read_raw import _get_supported as mne_get_supported


def open_raw(fname, preload=True, verbose="info"):
    """Reads a raw from file.

    Parameters
    ----------
    fname : str
        Path to the raw file.
    preload : bool
        Should the data be loaded as well or just the metadata.
    verbose : str
        Verbose level for the read_raw call.

    Returns
    -------
    mne.io.Raw
        The raw object read from the file.
    """
    try:
        if verbose == "info" or verbose == "debug":
            logging.getLogger("ui_logger").info("Reading " + fname)
            raw = mne.io.read_raw(fname, preload=preload)
        else:
            raw = mne.io.read_raw(fname, preload=preload, verbose=verbose)
        return raw
    except Exception:
        logging.getLogger("ui_logger").exception("")
        raise Exception("Could not read the raw file: " + str(fname))


def save_raw(raw, path, overwrite=True):
    """Saves a raw to file(s).

    For some safety, move the old files first to temp files,
    then save the new version, and then (if has not failed),
    remove the temp files.

    Parameters
    ----------
    raw : mne.io.Raw
        The raw file to be saved.
    path : str
        The path where to save.
    overwrite : bool
        Whether to overwrite.
    """

    logger = logging.getLogger("ui_logger")

    folder = os.path.dirname(path)
    bname = os.path.basename(path)

    if os.path.exists(path) and not overwrite:
        raise IOError("File already exists.")

    # Move existing files to temporary names
    ext = os.path.splitext(bname)[1]

    # Match filenames that
    # 1) do not start with an underscore,
    # 2) do not contain any extra dots, and
    # 3) end with the ext
    pat_old = re.compile(r"^(?!_)[^.]*" + re.escape(ext) + r"$")
    contents = os.listdir(folder)
    old_files = [fname_ for fname_ in contents if pat_old.match(fname_)]

    temp_paths = []
    for file_ in old_files:
        old_path = os.path.join(folder, os.path.basename(file_))
        temp_path = os.path.join(folder, "_" + os.path.basename(file_))
        logger.debug("Moving previously existing file to: " + str(temp_path))
        shutil.move(old_path, temp_path)
        temp_paths.append(temp_path)

    # Save raw data
    logger.debug("Saving new data to: " + str(path))
    raw.save(path, overwrite=True)

    # Remove old files
    for temp_path in temp_paths:
        logger.debug("Removing previously existing file: " + str(temp_path))
        os.remove(temp_path)

    # Just to make sure, set _filenames[0] to match the new path.
    raw._filenames = [path]


def ensure_folders(paths):
    """Ensures that paths in specified in the paths param exist.

    Parameters
    ----------
    paths : list
        List of folder paths.
    """
    for path in paths:
        if not os.path.exists(path):
            os.makedirs(path)


def save_csv(path, data, column_names, row_descs):
    """Saves tabular data to csv.

    Parameters
    ----------
    path : str
        Where to save.
    data : a numpy array of shape (n_rows, n_cols)
        Data to save.
    column_names : list
        List of column names.
    row_descs : list
        List of row descriptions that can be tuples like
        ('EEG', 'Left-frontal'), which are then put to the
        csv as multiple columns.
    """
    # gather all the data to list of rows
    all_data = []

    if isinstance(data, np.ndarray):
        data = data.tolist()

    # freqs data, assume same lengths
    all_data.append([""] * len(row_descs[0]) + column_names)

    for idx in range(len(data)):
        row = list(row_descs[idx]) + data[idx]
        all_data.append(row)

    # save to file
    all_data = np.array(all_data)
    np.savetxt(path, all_data, fmt="%s", delimiter=", ")


def load_csv(path):
    """Loads tabular data from csv.

    Parameters
    ----------
    path : str
        Path to the csv file.

    Returns
    -------
    list
        Column names.
    list
        Row descriptions.
    np.array
        The data.
    """

    all_data = np.loadtxt(
        path, dtype=str, delimiter=",", converters=lambda s: s.strip()
    )

    data = []
    column_names = []
    row_descs = []

    first_data_idx = np.min(np.where(all_data[0] != "")[0])

    column_names = all_data[0, first_data_idx:].tolist()
    row_descs = [tuple(elems) for elems in all_data[1:, :first_data_idx]]
    data = all_data[1:, first_data_idx:].astype(float)

    return column_names, row_descs, data


def get_resource_filename(package, resource_name):
    """
    Locate a file or folder inside a distribution package, supporting nested paths.

    Parameters
    ----------
    package : str
        Name of the distribution package to search.
    resource_name : str
        Relative path to the resource within the package.

    Returns
    -------
    str
        Absolute filesystem path to the resource.

    Raises
    ------
    FileNotFoundError
        If the resource cannot be found in the given package.
    """
    parts = os.path.normpath(resource_name).split(os.sep)
    try:
        resource_path = _resources.files(package).joinpath(*parts)
        with _resources.as_file(resource_path) as p:
            return str(p)
    except Exception as e:
        raise FileNotFoundError(
            f"Resource {resource_name!r} not found in package {package!r}: {e}"
        )


def get_distribution_version(dist_name):
    """
    Return the installed version of a distribution, or empty string if missing.

    Parameters
    ----------
    dist_name : str
        Name of the distribution package to query.

    Returns
    -------
    str
        Version string, or '' if the package is not installed.
    """
    try:
        return _metadata.version(dist_name)
    except _metadata.PackageNotFoundError:
        return ""


def get_package_keys():
    """
    Return all installed distribution package names as lowercase with dashes replaced by underscores.

    Returns
    -------
    list of str
        Installed package names, lowercased and with '-' converted to '_'.
    """
    keys = []
    for dist in _metadata.distributions():
        name = dist.metadata.get("Name", "")
        if name:
            keys.append(name.lower().replace("-", "_"))
    return keys


def homepath():
    """Tries to find correct path for home folder.

    Returns
    -------
    str
        Path to home directory.

    """

    home = os.path.expanduser("~")

    if not home:
        return "."

    return home


def configpath():
    """Find a path for configuration files.

    Returns
    -------
    str
        Path to config directory

    """
    return appdirs.user_config_dir("meggie")


def datapath():
    """Find a path for data files.

    Returns
    -------
    str
        Path to data directory

    """
    return os.path.join(appdirs.user_data_dir("meggie"), "experiments")


def get_supported_formats():
    """
    Get all raw‐file formats supported by MNE.

    Returns
    -------
    list of tuple
        Each tuple is (file extension, [reader names]) supported by MNE.
    """
    mne_supported = mne_get_supported()

    items = []
    for ext_item in mne_supported.items():
        ext = ext_item[0]
        names = [val[0] for val in ext_item[1].items()]
        items.append((ext, names))

    return items



"""Functions for validation."""

import re
import numpy as np


def validate_name(name, minlength=1, maxlength=30, fieldname="name"):
    """Validates a name with length and regular expression criteria.

    Parameters
    ----------
    name : str or QString
        Name to be validated.
    minlength : int
        The minimum length of the name.
    maxlength : int
        The maximum length of the name.
    fieldname : str
        The name of the field, for good exception messages.

    Returns
    -------
    str
        Name that passed the validation.

    """

    name = str(name)

    if len(name) < minlength:
        raise Exception("You need to set " + fieldname)

    if len(name) > maxlength:
        raise Exception(
            "Too long " + fieldname + " (over " + str(maxlength) + " characters.)"
        )

    if not re.match(r"^[A-Za-z0-9_]*$", name):
        raise Exception(
            fieldname + " can only contain alphanumeric " + "characters or underscores"
        )

    return name


def assert_arrays_same(arrays, message="Times do not match"):
    """Checks if arrays in a list are all equal.

    Parameters
    ----------
    arrays : list
        List containing the numpy arrays to be compared.
    message : str
        Message for the exception if arrays are not the same.
    """
    for idx in range(len(arrays) - 1):
        try:
            first = np.array(arrays[0]).astype(float)
            second = np.array(arrays[idx + 1]).astype(float)
            np.testing.assert_array_almost_equal(first, second)
        except Exception:
            raise Exception(message)


def assert_lists_same(lists, message):
    """Checks if lists in a list are all equal.

    Parameters
    ----------
    lists : list
        List containing lists to be compared.
    message : str
        Message for the exception if lists are not the same.
    """
    try:
        sets = [set(lst) for lst in lists]
        assert all(e == sets[0] for e in sets)
    except Exception:
        raise Exception(message)



"""Contains implementation for raw montage"""

import mne

from meggie.utilities.threading import threaded

from meggie.actions.raw_montage.dialogs.montageDialogMain import MontageDialog

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class Montage(Action):
    """Shows a dialog for gathering parameters and then
    allows setting montage for EEG.
    """

    def run(self, params={}):
        montage_dialog = MontageDialog(self.window, self.experiment, self.handler)
        montage_dialog.show()

    @subject_action
    def handler(self, subject, params):
        """ """

        @threaded
        def montage_fun():
            """ """
            head_size = params["head_size"]

            if params["custom"] is True:
                montage_fname = params["selection"]
                montage = mne.channels.read_custom_montage(
                    montage_fname, head_size=head_size
                )
            else:
                montage_name = params["selection"]
                montage = mne.channels.make_standard_montage(
                    montage_name, head_size=head_size
                )

            if subject.has_raw:
                raw = subject.get_raw()
                raw.set_montage(montage)

            subject.save_montage(montage)
            subject.save()

        montage_fun(do_meanwhile=self.window.update_ui)



from meggie.utilities.testing import BaseTestAction
from meggie.actions.spectrum_delete_from_all import DeleteSpectrumFromAll


class TestSpectrumDeleteFromAll(BaseTestAction):
    def test_spectrum_delete_from_all(self):

        data = {"outputs": {"spectrum": ["Spectrum"]}}

        self.run_action(
            action_name="spectrum_delete_from_all",
            handler=DeleteSpectrumFromAll,
            data=data,
        )



from meggie.utilities.testing import BaseTestAction
from meggie.actions.evoked_info import Info


class TestEvokedInfo(BaseTestAction):
    def test_evoked_info(self):

        data = {"outputs": {"evoked": ["Evoked"]}}

        content = self.run_action(
            action_name="evoked_info",
            handler=Info,
            data=data,
        )
        assert content



"""Contains save tse action handling."""

import os

from PyQt5 import QtWidgets

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.validators import assert_arrays_same
from meggie.utilities.filemanager import homepath

from meggie.mainwindow.dynamic import Action

from meggie.utilities.dialogs.TFROutputOptionsMain import TFROutputOptions

from meggie.actions.tfr_save_tse.controller.tfr import save_tse_channel_averages
from meggie.actions.tfr_save_tse.controller.tfr import save_tse_all_channels


class SaveTSE(Action):
    """Saves TSE's to csv files"""

    def run(self, params={}):
        try:
            selected_name = self.data["outputs"]["tfr"][0]
        except IndexError:
            return

        time_arrays = []
        freq_arrays = []
        for subject in self.experiment.subjects.values():
            tfr = subject.tfr.get(selected_name)
            if not tfr:
                continue
            time_arrays.append(tfr.times)
            freq_arrays.append(tfr.freqs)
        assert_arrays_same(time_arrays)
        assert_arrays_same(freq_arrays, "Freqs do no match")

        def option_handler(params):

            default_filename = (
                selected_name + "_all_subjects_channel_averages_tse.csv"
                if params["output_option"] == "channel_averages"
                else selected_name + "_all_subjects_all_channels_tse.csv"
            )
            filepath, _ = QtWidgets.QFileDialog.getSaveFileName(
                self.window,
                "Save TSE to CSV",
                os.path.join(homepath(), default_filename),
                "CSV Files (*.csv);;All Files (*)",
            )

            if not filepath:
                return

            params["channel_groups"] = self.experiment.channel_groups
            params["name"] = selected_name
            params["filepath"] = filepath

            try:
                self.handler(self.experiment.active_subject, params)
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = TFROutputOptions(
            self.window, self.experiment, selected_name, handler=option_handler
        )
        dialog.show()

    def handler(self, subject, params):
        """ """
        if params["output_option"] == "all_channels":
            save_tse_all_channels(
                self.experiment,
                params["name"],
                params["blmode"],
                params["blstart"],
                params["blend"],
                params["tmin"],
                params["tmax"],
                params["fmin"],
                params["fmax"],
                params["filepath"],
                do_meanwhile=self.window.update_ui,
            )
        else:
            save_tse_channel_averages(
                self.experiment,
                params["name"],
                params["blmode"],
                params["blstart"],
                params["blend"],
                params["tmin"],
                params["tmax"],
                params["fmin"],
                params["fmax"],
                params["channel_groups"],
                params["filepath"],
                do_meanwhile=self.window.update_ui,
            )



from meggie.utilities.testing import BaseTestAction
from meggie.actions.spectrum_save import SaveSpectrum
from meggie.utilities.dialogs.outputOptionsMain import OutputOptions


class TestSpectrumSaveChannelAverages(BaseTestAction):
    def test_spectrum_save_channel_averages(self):

        data = {"outputs": {"spectrum": ["Spectrum"]}}

        self.run_action(
            action_name="spectrum_save",
            handler=SaveSpectrum,
            data=data,
        )
        dialog = self.find_dialog(OutputOptions)
        dialog.ui.radioButtonChannelAverages.setChecked(True)
        dialog.accept()



"""Contains create epochs action handling."""

from meggie.utilities.names import next_available_name
from meggie.utilities.threading import threaded

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.actions.epochs_create.dialogs.createEpochsFromEventsDialogMain import (
    CreateEpochsFromEventsDialog,
)
from meggie.actions.epochs_create.controller.epoching import create_epochs_from_events


class CreateEpochs(Action):
    """Shows a dialog for event selection and then creates epochs."""

    def run(self, params={}):
        default_name = next_available_name(
            self.experiment.active_subject.epochs.keys(), "Epochs"
        )
        dialog = CreateEpochsFromEventsDialog(
            self.experiment, self.window, default_name, self.handler
        )
        dialog.show()

    @subject_action
    def handler(self, subject, params):
        """ """

        @threaded
        def create():
            create_epochs_from_events(subject, params)

        create(do_meanwhile=self.window.update_ui)
        self.experiment.save_experiment_settings()



import os

import mne

from meggie.utilities.measurement_info import MeasurementInfo


def test_measurement_info():
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, "MEG", "sample", "sample_audvis_raw.fif")
    raw = mne.io.read_raw_fif(sample_fname, preload=True)

    meas_info = MeasurementInfo(raw)

    assert meas_info.high_pass == 0.1
    assert meas_info.low_pass == 172.18
    assert meas_info.sampling_freq == 600.61
    assert meas_info.date == "2002-12-03"
    assert meas_info.subject_name == ""



"""Contains logic for the epoch creation dialog."""

import logging

from PyQt5 import QtWidgets

from meggie.actions.epochs_create.dialogs.createEpochsFromEventsDialogUi import (
    Ui_CreateEpochsFromEventsDialog,
)

from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget
from meggie.utilities.dialogs.bitSelectionDialogMain import BitSelectionDialog

from meggie.utilities.validators import validate_name
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox


class CreateEpochsFromEventsDialog(QtWidgets.QDialog):
    """Dialog responsible for collecting parameters for epoch creation."""

    def __init__(self, experiment, parent, default_name, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_CreateEpochsFromEventsDialog()
        self.ui.setupUi(self)

        self.parent = parent
        self.experiment = experiment
        self.handler = handler

        self.events = []

        self.batching_widget = BatchingWidget(
            experiment_getter=self._experiment_getter,
            parent=self,
            container=self.ui.groupBoxBatching,
            geometry=self.ui.batchingWidgetPlaceholder.geometry(),
        )
        self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)

        self.ui.lineEditCollectionName.setText(default_name)

    def _experiment_getter(self):
        return self.experiment

    def _update_events(self):
        self.ui.listWidgetEvents.clear()

        events = self.events

        for event in events:
            item = QtWidgets.QListWidgetItem(
                "%s, %s"
                % ("ID " + str(event["event_id"]), "mask=" + str(event["mask"]))
            )
            self.ui.listWidgetEvents.addItem(item)

    def _collect_parameter_values(self):
        tmin = float(self.ui.doubleSpinBoxTmin.value())
        tmax = float(self.ui.doubleSpinBoxTmax.value())
        bstart = float(self.ui.doubleSpinBoxBaselineStart.value())
        bend = float(self.ui.doubleSpinBoxBaselineEnd.value())
        delay = float(self.ui.doubleSpinBoxDelay.value())

        mag = self.ui.checkBoxMag.isChecked()
        grad = self.ui.checkBoxGrad.isChecked()
        eeg = self.ui.checkBoxEeg.isChecked()

        collection_name = validate_name(str(self.ui.lineEditCollectionName.text()))

        reject = dict()
        if mag:
            reject["mag"] = self.ui.doubleSpinBoxMagReject.value()
        if grad:
            reject["grad"] = self.ui.doubleSpinBoxGradReject.value()
        if eeg:
            reject["eeg"] = self.ui.doubleSpinBoxEEGReject.value()

        params = {
            "mag": mag,
            "grad": grad,
            "eeg": eeg,
            "reject": reject,
            "tmin": tmin,
            "tmax": tmax,
            "bstart": bstart,
            "bend": bend,
            "delay": delay,
            "collection_name": collection_name,
            "events": self.events,
        }
        return params

    def accept(self):
        if self.ui.listWidgetEvents.count() == 0:
            message = "Cannot create epochs from empty list."
            messagebox(self.parent, message)
            return

        try:
            params = self._collect_parameter_values()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        subject = self.experiment.active_subject

        try:
            self.handler(subject, params)
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.initialize_ui()

        self.close()

    def acceptBatch(self):

        if self.ui.listWidgetEvents.count() == 0:
            message = "Cannot create epochs from empty list."
            messagebox(self.parent, message)
            return

        try:
            params = self._collect_parameter_values()
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        params["run_in_batch"] = True

        selected_subject_names = self.batching_widget.selected_subjects
        for name, subject in self.experiment.subjects.items():
            if name in selected_subject_names:
                try:
                    self.handler(subject, params)
                    subject.release_memory()
                except Exception as exc:
                    self.batching_widget.failed_subjects.append((subject, str(exc)))
                    logging.getLogger("ui_logger").exception("")

        self.batching_widget.cleanup()

        self.parent.initialize_ui()
        self.close()

    def on_pushButtonEdit_clicked(self, checked=None):
        if checked is None:
            return

        self.bitDialog = BitSelectionDialog(
            self, self.ui.spinBoxMask, self.ui.spinBoxEventID
        )

        self.bitDialog.show()

    def on_pushButtonHelp_clicked(self, checked=None):
        if checked is None:
            return

        help_message = (
            "Events are found in a following way. If only event "
            "id is set, events with exactly the same binary representation as "
            "event id are included in the final event list. If also mask is "
            "set, event list will also include events where binary digits in "
            "the places specified by the mask are not the same as in the "
            "event id, or in other words, only events where the digits we "
            "are interested in are the same as in the list of all events, "
            "are included. Binary representations are assumed to be 16 digits "
            "long. \n\nFor example event id of 0000010000010000 = 1040 and "
            "mask of 0000000000000011 = 3 would mean that first (rightmost) "
            "two digits can be 1 or 0, but anything else must be exactly as "
            "in the event id. Thus events with following id's would be allowed:"
            "\n\n0000010000010000 = 1040\n0000010000010001 = 1041\n"
            "0000010000010010 = 1042\n0000010000010011 = 1043"
        )

        messagebox(self.parent, help_message, "Mask help")

    def on_pushButtonAdd_clicked(self, checked=None):
        if checked is None:
            return

        event_params = {
            "mask": self.ui.spinBoxMask.value(),
            "event_id": self.ui.spinBoxEventID.value(),
        }

        if event_params not in self.events:
            self.events.append(event_params)
            self._update_events()

    def on_pushButtonClear_clicked(self, checked=None):
        if checked is None:
            return

        self.events = []
        self._update_events()



from meggie.utilities.testing import BaseTestAction
from meggie.actions.tfr_save_tse import SaveTSE
from meggie.utilities.dialogs.TFROutputOptionsMain import TFROutputOptions


class TestTFRSaveTSEAllChannels(BaseTestAction):
    def test_tfr_save_tse_all_channels(self):

        data = {"outputs": {"tfr": ["TFR"]}}

        self.run_action(
            action_name="tfr_save_tse",
            handler=SaveTSE,
            data=data,
        )
        dialog = self.find_dialog(TFROutputOptions)
        dialog.ui.radioButtonChannelAverages.setChecked(False)
        dialog.accept()



from meggie.utilities.testing import BaseTestAction
from meggie.actions.spectrum_plot import PlotSpectrum
from meggie.utilities.dialogs.outputOptionsMain import OutputOptions


class TestSpectrumPlotAllChannels(BaseTestAction):
    def test_spectrum_plot_all_channels(self):

        data = {"outputs": {"spectrum": ["Spectrum"]}}

        self.run_action(
            action_name="spectrum_plot",
            handler=PlotSpectrum,
            data=data,
        )
        dialog = self.find_dialog(OutputOptions)
        dialog.ui.radioButtonChannelAverages.setChecked(False)
        dialog.accept()



"""Contains implementation for epochs plot"""

import logging

import matplotlib.pyplot as plt

from meggie.utilities.messaging import exc_messagebox
from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class PlotEpochs(Action):
    """Plots all epochs from all channels."""

    def run(self, params={}):

        try:
            selected_name = self.data["outputs"]["epochs"][0]
        except IndexError:
            return

        subject = self.experiment.active_subject
        epochs = subject.epochs.get(selected_name)

        # Create a copy to not drop epochs from the original yet
        mne_epochs = epochs.content.copy()

        old_selection = mne_epochs.selection.copy()

        def handle_close(event):
            selection_changed = len(old_selection) != len(mne_epochs.selection)

            params = {}
            params["name"] = selected_name
            params["selection_changed"] = selection_changed
            params["selection"] = mne_epochs.selection

            try:
                if selection_changed:
                    self.handler(subject, params)
                    self.window.initialize_ui()
            except Exception as exc:
                exc_messagebox(self.window, exc)

        fig = mne_epochs.plot(show=False)
        fig.canvas.mpl_connect("close_event", handle_close)
        plt.show()

    @subject_action
    def handler(self, subject, params):

        logging.getLogger("ui_logger").info("Epochs dropped!")

        epochs = subject.epochs.get(params["name"])

        # Finally, drop the epochs
        mne_epochs = epochs.content
        dropped = [
            sel_idx
            for sel_idx in mne_epochs.selection
            if sel_idx not in params["selection"]
        ]
        dropped_idxs = [
            idx
            for idx, sel_idx in enumerate(mne_epochs.selection)
            if sel_idx in dropped
        ]
        mne_epochs.drop(dropped_idxs)

        # And save
        epochs.save_content()



from meggie.utilities.testing import BaseTestAction
from meggie.actions.tfr_delete import DeleteTFR


class TestTFRDelete(BaseTestAction):
    def test_tfr_delete(self):

        data = {"outputs": {"tfr": ["TFR"]}}

        self.run_action(
            action_name="tfr_delete",
            handler=DeleteTFR,
            data=data,
        )



"""Contains implementation for raw plot projections"""

from meggie.utilities.messaging import messagebox
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.plotting import set_figure_title

from meggie.mainwindow.dynamic import Action


class PlotProjections(Action):
    """Shows a plot of existing projection vectors."""

    def run(self, params={}):

        subject = self.experiment.active_subject
        try:
            self.handler(subject, {})
        except Exception as exc:
            exc_messagebox(self.window, exc)

    def handler(self, subject, params):
        raw = subject.get_raw()
        if not raw.info["projs"]:
            messagebox(self.window, "Data does not contain any projection vectors")
            return
        fig = raw.plot_projs_topomap()

        elems = ["projections", subject.name]
        set_figure_title(fig, "_".join(elems))
        fig.suptitle(" ".join(elems))



import mne

from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_events_from_annotations import EventsFromAnnotations
from meggie.actions.raw_events_from_annotations.dialogs.eventsFromAnnotationsDialogMain import (
    EventsFromAnnotationsDialog,
)


class TestRawEventsFromAnnotations(BaseTestAction):

    def setup_experiment(self):
        BaseTestAction.setup_experiment(self)
        raw = self.experiment.active_subject.get_raw()

        # create one annotation named "CAT"
        annotations = mne.Annotations([2], [1], ["CAT"])
        raw.set_annotations(annotations)

    def test_raw_events_from_annotations(self):
        self.run_action(
            action_name="raw_events_from_annotations",
            handler=EventsFromAnnotations,
            patch_paths=[
                "meggie.actions.raw_events_from_annotations.dialogs.eventsFromAnnotationsDialogMain"
            ],
        )
        dialog = self.find_dialog(EventsFromAnnotationsDialog)
        dialog.items = [("CAT", 99, True)]
        dialog.accept()



from meggie.utilities.testing import BaseTestAction
from meggie.actions.epochs_plot import PlotEpochs


class TestEpochsPlot(BaseTestAction):
    def test_epochs_plot(self):

        data = {"outputs": {"epochs": ["Epochs"]}}

        self.run_action(action_name="epochs_plot", handler=PlotEpochs, data=data)



"""Contains controlling logic for the evoked plot"""

import mne
import matplotlib.pyplot as plt

from matplotlib.lines import Line2D

from meggie.utilities.plotting import color_cycle
from meggie.utilities.plotting import create_channel_average_plot
from meggie.utilities.channels import average_to_channel_groups
from meggie.utilities.channels import ensure_montage
from meggie.utilities.plotting import set_figure_title

from meggie.utilities.units import get_unit


def _create_averages(mne_evoked, channel_groups):
    mne_evoked = mne_evoked.copy().drop_channels(mne_evoked.info["bads"])

    data_labels, averaged_data = average_to_channel_groups(
        mne_evoked.data, mne_evoked.info, mne_evoked.info["ch_names"], channel_groups
    )

    return data_labels, averaged_data


def plot_evoked_averages(evoked, channel_groups):
    """Plots channel averages."""
    conditions = evoked.content.keys()
    colors = color_cycle(len(conditions))
    times = evoked.times

    averages = {}
    for key, mne_evoked in sorted(evoked.content.items()):
        data_labels, averaged_data = _create_averages(mne_evoked, channel_groups)

        for label_idx, label in enumerate(data_labels):
            if label not in averages:
                averages[label] = []
            averages[label].append((key, averaged_data[label_idx]))

    ch_types = sorted(set([label[0] for label in averages.keys()]))
    for ch_type in ch_types:

        ch_groups = sorted(
            [label[1] for label in averages.keys() if label[0] == ch_type]
        )

        def plot_fun(ax_idx, ax):
            ch_group = ch_groups[ax_idx]
            ax.set_title(ch_group)

            ax.set_xlabel("Time (s)")
            ax.set_ylabel("Amplitude ({})".format(get_unit(ch_type)))

            for color_idx, (key, curve) in enumerate(averages[(ch_type, ch_group)]):
                ax.plot(times, curve, color=colors[color_idx])

            ax.axhline(0, color="black")
            ax.axvline(0, color="black")

        title = " ".join([evoked.name, ch_type])
        legend = list(zip(conditions, colors))
        create_channel_average_plot(len(ch_groups), plot_fun, title, legend)

    plt.show()


def plot_evoked_topo(subject, evoked, ch_type):
    """Plots evoked time courses arranged as a topography"""
    evokeds = []
    labels = []
    for key, mne_evoked in sorted(evoked.content.items()):

        info = mne_evoked.info
        if ch_type == "eeg":
            dropped_names = [
                ch_name
                for ch_idx, ch_name in enumerate(info["ch_names"])
                if ch_idx not in mne.pick_types(info, eeg=True, meg=False)
            ]
        else:
            dropped_names = [
                ch_name
                for ch_idx, ch_name in enumerate(info["ch_names"])
                if ch_idx not in mne.pick_types(info, eeg=False, meg=True)
            ]

        mne_evoked = mne_evoked.copy().drop_channels(dropped_names)

        ensure_montage(subject, mne_evoked, ch_type)

        evokeds.append(mne_evoked)

        labels.append(key)

    colors = color_cycle(len(evoked.content.keys()))

    # setup legend for subplots
    lines = [
        Line2D([0], [0], color=colors[idx], label=labels[idx])
        for idx in range(len(labels))
    ]

    fig, axes = plt.subplots()

    mne.viz.plot_evoked_topo(evokeds, color=colors, legend=False, show=False, axes=axes)
    axes.legend(handles=lines, loc="upper right")
    title = "{0}_{1}".format(evoked.name, ch_type)
    set_figure_title(fig, title)

    def onclick(event):
        try:
            # not nice:
            ax = plt.gca()

            if ax is axes:
                return

            channel = plt.getp(ax, "title")
            ax.set_title("")

            ax.legend(handles=lines, loc="upper right")

            title = " ".join([evoked.name, channel])
            ax.figure.suptitle(title)
            set_figure_title(ax.figure, title.replace(" ", "_"))
            plt.show()
        except Exception:
            pass

    fig.canvas.mpl_connect("button_press_event", onclick)

    plt.show()



import tempfile
import logging
import importlib
import matplotlib
import pytest
import json
import os
from meggie.utilities.filemanager import get_resource_filename
import mne
import shutil
import numpy as np
import PyQt5.QtWidgets
from PyQt5.QtWidgets import QApplication, QMainWindow
from meggie.mainwindow.preferences import PreferencesHandler
from meggie.experiment import initialize_new_experiment
from meggie.utilities.events import find_events
from meggie.utilities.events import find_stim_channel
from meggie.datatypes.epochs.epochs import Epochs
from meggie.datatypes.evoked.evoked import Evoked
from meggie.datatypes.spectrum.spectrum import Spectrum
from meggie.datatypes.tfr.tfr import TFR

os.environ["QT_QPA_PLATFORM"] = "offscreen"
mne.viz.set_browser_backend("matplotlib")
matplotlib.use("Agg")


def create_multi_sample_experiment(
    experiment_folder, experiment_name, subjects_raw, overwrite=False
):
    """Creates experiment from list of raw objects.

    Parameters
    ----------
    experiment_folder : str
        Where the experiment is stored.
    experiment_name : str
        Name of the experiment.
    subjects_raw : list
        List of raws, one for each subject.
    overwrite : bool
        Whether to overwrite previous experiment with a same name.

    Returns
    -------
    meggie.experiment.Experiment
        An experiment object containing the subjects.
    """

    if os.path.exists(os.path.join(experiment_folder, experiment_name)):
        if overwrite:
            shutil.rmtree(os.path.join(experiment_folder, experiment_name))
        else:
            raise Exception("Experiment already exists")

    # Create preferences object to store working directory
    prefs = PreferencesHandler()
    prefs.workspace = experiment_folder

    # create experiment (creates experiment directory inside working directory)
    experiment = initialize_new_experiment(
        experiment_name, "Test", prefs, set_previous_experiment=False
    )

    for subject_idx, raw in enumerate(subjects_raw):
        sample_name = "sample_" + str(subject_idx + 1).zfill(2) + "-raw"
        sample_folder = os.path.join(experiment_folder, experiment_name, sample_name)
        os.makedirs(sample_folder)
        fname = sample_name + ".fif"

        raw_path = os.path.join(sample_folder, fname)
        raw.save(raw_path)

        experiment.create_subject(sample_name, raw_path)

    experiment.save_experiment_settings()
    return experiment


def create_evoked_conditions_experiment(
    experiment_folder, experiment_name, overwrite=False, n_subjects=35
):
    """Generate multisubject experiment based on sample_audvis_raw for testing purposes.

    Subjects are generated so that each will contain different versions of two LA and RA auditory responses.

    Parameters
    ----------
    experiment_folder : str
        Where the experiment is stored.
    experiment_name : str
        Name of the experiment.
    overwrite : bool
        Whether to overwrite previous experiment with a same name.
    n_subjects : int
        How many subjects to add.

    Returns
    -------
    meggie.experiment.Experiment
        An experiment object containing the subjects.
    """
    sample_folder = mne.datasets.sample.data_path()
    sample_fname = os.path.join(sample_folder, "MEG", "sample", "sample_audvis_raw.fif")

    subjects_raw = []

    # generate data
    raw = mne.io.read_raw_fif(sample_fname, preload=True)
    events = mne.find_events(raw)

    la_events = [ev for ev in events if ev[2] == 1]
    ra_events = [ev for ev in events if ev[2] == 2]

    for subject_idx in range(n_subjects):
        data = []
        combined_events = la_events[subject_idx * 2 : subject_idx * 2 + 2]
        combined_events.extend(ra_events[subject_idx * 2 : subject_idx * 2 + 2])

        if len(combined_events) != 4:
            raise Exception("Something wrong with event counts")

        data = []
        for event in combined_events:
            tidx = event[0]
            try:
                ev_before = [ev for ev in events if ev[0] < tidx][-1]
                tmin = (ev_before[0] - raw.first_samp + 20) / raw.info["sfreq"]
            except Exception:
                tmin = 0.0

            try:
                ev_after = [ev for ev in events if ev[0] > tidx][0]
                tmax = (ev_after[0] - raw.first_samp - 20) / raw.info["sfreq"]
            except Exception:
                tmax = None

            data.append(raw.copy().crop(tmin, tmax)._data)

        subject_raw = mne.io.RawArray(
            np.concatenate(data, axis=1), raw.info, first_samp=0
        )

        subjects_raw.append(subject_raw)

    return create_multi_sample_experiment(
        experiment_folder, experiment_name, subjects_raw, overwrite=overwrite
    )


def create_test_experiment(experiment_folder, experiment_name, n_subjects=2):
    """Generate experiment with data for testing."""

    experiment = create_evoked_conditions_experiment(
        experiment_folder, experiment_name, n_subjects=n_subjects
    )

    # create trivial content
    for subject in experiment.subjects.values():
        raw = subject.get_raw()

        # create epochs
        stim_channel = find_stim_channel(raw)
        events = find_events(raw, stim_channel)

        params = {
            "tmin": -0.1,
            "tmax": 0.4,
            "bstart": -0.1,
            "bend": 0.0,
        }
        category = {"1": 1}
        mne_epochs = mne.Epochs(
            raw,
            events,
            category,
            tmin=params["tmin"],
            tmax=params["tmax"],
            baseline=(params["bstart"], params["bend"]),
        )
        epochs_directory = subject.epochs_directory
        epochs = Epochs("Epochs", epochs_directory, params, content=mne_epochs)
        epochs.save_content()
        subject.add(epochs, "epochs")

        params = {
            "tmin": -0.1,
            "tmax": 0.4,
            "bstart": -0.1,
            "bend": 0.0,
        }
        category = {"2": 2}
        mne_epochs_2 = mne.Epochs(
            raw,
            events,
            category,
            tmin=params["tmin"],
            tmax=params["tmax"],
            baseline=(params["bstart"], params["bend"]),
        )
        epochs_directory = subject.epochs_directory
        epochs_2 = Epochs("Epochs2", epochs_directory, params, content=mne_epochs_2)
        epochs_2.save_content()
        subject.add(epochs_2, "epochs")

        # create evoked
        mne_evoked = mne_epochs.average()
        mne_evoked_2 = mne_epochs_2.average()
        params = {"conditions": ["Epochs", "Epochs2"]}
        content = {"Epochs": mne_evoked, "Epochs2": mne_evoked_2}
        evoked_directory = subject.evoked_directory
        evoked = Evoked("Evoked", evoked_directory, params, content=content)
        evoked.save_content()
        subject.add(evoked, "evoked")

        # create spectrum
        picks = mne.pick_types(raw.info, meg=True, eeg=True)
        info = mne.pick_info(raw.info, sel=picks)

        fmin = 1
        fmax = 40
        nfft = 64
        overlap = 32

        tmin = 0
        tmax = raw.times[-1] / 2
        mne_spectrum = raw.compute_psd(
            method="welch",
            fmin=fmin,
            fmax=fmax,
            tmin=tmin,
            tmax=tmax,
            n_fft=nfft,
            n_overlap=overlap,
            picks=picks,
        )
        psds = mne_spectrum.get_data()

        tmin_2 = raw.times[-1] / 2
        tmax_2 = raw.times[-1]
        mne_spectrum_2 = raw.compute_psd(
            method="welch",
            fmin=fmin,
            fmax=fmax,
            tmin=tmin_2,
            tmax=tmax_2,
            n_fft=nfft,
            n_overlap=overlap,
            picks=picks,
        )
        psds_2 = mne_spectrum_2.get_data()

        psd_data = {"1": psds, "2": psds_2}
        freqs = mne_spectrum.freqs

        params = {
            "fmin": fmin,
            "fmax": fmax,
            "nfft": nfft,
            "overlap": overlap,
            "conditions": ["1", "2"],
            "intervals": {"1": [(tmin, tmax)], "2": [(tmin_2, tmax_2)]},
        }
        spectrum = Spectrum(
            "Spectrum", subject.spectrum_directory, params, psd_data, freqs, info
        )
        spectrum.save_content()
        subject.add(spectrum, "spectrum")

        # create tfr
        minfreq = 20
        maxfreq = 40
        interval = 5
        n_cycles = 1
        decim = 1
        freqs = np.arange(minfreq, maxfreq, interval)

        mne_tfr = mne_epochs.compute_tfr(
            "morlet",
            freqs,
            n_cycles=n_cycles,
            decim=decim,
            average=True,
            return_itc=False,
        )
        mne_tfr_2 = mne_epochs_2.compute_tfr(
            "morlet",
            freqs,
            n_cycles=n_cycles,
            decim=decim,
            average=True,
            return_itc=False,
        )
        tfr_data = {"Epochs": mne_tfr, "Epochs2": mne_tfr_2}

        params = {
            "decim": decim,
            "n_cycles": n_cycles,
            "evoked_subtracted": False,
            "conditions": ["Epochs", "Epochs2"],
        }
        tfr = TFR("TFR", subject.tfr_directory, params, tfr_data)
        tfr.save_content()
        subject.add(tfr, "tfr")

    return experiment


class MockMainWindow(QMainWindow):
    def __init__(self, *args, **kwargs):
        self.ui = None
        self.prefs = PreferencesHandler()
        QMainWindow.__init__(self)

    def update_ui(self):
        pass

    def initialize_ui(self):
        pass


def patched_messagebox(parent, message):
    raise Exception(message)


def patched_exc_messagebox(parent, exc):
    raise exc


def patched_logger_exception(msg):
    raise Exception(msg)


class BaseTestAction:
    @pytest.fixture(autouse=True)
    def setup_common(self, qtbot, monkeypatch):
        # before each test
        os.environ["QT_QPA_PLATFORM"] = "offscreen"
        self.qtbot = qtbot
        self.monkeypatch = monkeypatch

        # pyqt6 insists of having QApplication before creating the main window
        self.ensure_app()

        self.mock_main_window = MockMainWindow()
        self.temp_dir = tempfile.TemporaryDirectory()
        self.dirpath = self.temp_dir.name
        self.setup_experiment()
        yield

        self.temp_dir.cleanup()

    def setup_experiment(self):
        self.experiment = create_test_experiment(self.dirpath, "test_experiment")
        self.experiment.activate_subject("sample_01-raw")

    def run_action(self, action_name, handler, data={}, patch_paths=[]):
        # patch logger to raise exceptions
        logger = logging.getLogger("ui_logger")
        self.monkeypatch.setattr(logger, "exception", patched_logger_exception)

        basepath = f"meggie.actions.{action_name}"
        if basepath not in patch_paths:
            patch_paths.append(basepath)

        # patch messageboxes to raise exceptions
        for patch_path in patch_paths:
            try:
                module = importlib.import_module(patch_path)
            except ModuleNotFoundError:
                continue

            if getattr(module, "exc_messagebox", None):
                self.monkeypatch.setattr(
                    ".".join([patch_path, "exc_messagebox"]),
                    patched_exc_messagebox,
                )

            if getattr(module, "messagebox", None):
                self.monkeypatch.setattr(
                    ".".join([patch_path, "messagebox"]),
                    patched_messagebox,
                )

        # mock savefile-dialog to allow tests pass
        def mocked_save_file(parent, title, path, suffix):
            return f"{self.dirpath}/results.csv", None

        PyQt5.QtWidgets.QFileDialog.getSaveFileName = mocked_save_file

        # call the action handler
        data = data.copy()
        data.update({"tab_id": "test_tab"})
        action_spec = self.load_action_spec(action_name)
        self.action_instance = handler(
            self.experiment, data, self.mock_main_window, action_spec
        )
        return self.action_instance.run()

    def load_action_spec(self, action_name):
        package_name = self.__class__.__module__.split(".")[0]
        action_path = get_resource_filename(package_name, "actions")
        config_path = os.path.join(action_path, action_name, "configuration.json")
        with open(config_path, "r") as f:
            action_spec = json.load(f)
        return action_spec

    def find_dialog(self, dialog_class):
        count = 0
        for widget in QApplication.topLevelWidgets():
            if isinstance(widget, dialog_class):
                # check that the dialog isn't closed
                if not widget.isVisible():
                    continue
                count += 1
                dialog = widget
        assert count == 1

        assert dialog is not None
        return dialog

    def ensure_app(self):
        self.app = QApplication.instance()
        if not self.app:
            self.app = QApplication([])



"""Provides functions for popping up different kind of messageboxes."""

import logging
from meggie.utilities.dialogs.shortMessageBoxMain import shortMessageBox
from meggie.utilities.dialogs.shortQuestionBoxMain import shortQuestionBox


def exc_messagebox(parent, exc):
    """Pops up a messagebox for a exception.

    Parameters
    ----------
    parent : QDialog or QMainWindow
        Parent dialog, that is set as a parent.
    exc : instance of Exception
        The exception that is presented.
    """

    try:
        if isinstance(exc, str):
            error_message = exc
        else:
            error_message = str(exc)
            if not error_message:
                error_message = "(empty message of type " + str(exc.__repr__()) + ")"
    except Exception:
        error_message = ""

    # print traceback to console
    logging.getLogger("ui_logger").exception("")

    # create messagebox for user
    message = "\n\n".join(
        [
            "There has been an error with following message: ",
            error_message,
            "More information can be found from console below.",
        ]
    )

    messagebox = shortMessageBox(message, parent)
    messagebox.show()


def messagebox(parent, msg):
    """Pops up a messagebox with a message.

    Parameters
    ----------
    parent : QDialog or QMainWindow
        Parent dialog, that is set as a parent.
    msg : str
        The message that is presented.
    """
    messagebox = shortMessageBox(msg, parent)
    messagebox.show()


def questionbox(parent, question, handler):
    """Pops up a yes/no questionbox.

    Parameters
    ----------
    parent : QDialog or QMainWindow
        Parent dialog, that is set as a parent.
    question : str
        The question that is asked.
    handler : function
        A function that is called with the answer used provided.

    """

    questionbox = shortQuestionBox(question, parent, handler)
    questionbox.exec()



from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_ica import ICA
from meggie.actions.raw_ica.dialogs.icaDialogMain import (
    ICADialog,
)


class TestRawICA(BaseTestAction):
    def test_raw_ica(self):

        self.run_action(
            action_name="raw_ica",
            handler=ICA,
            patch_paths=["meggie.actions.raw_ica.dialogs.icaDialogMain"],
        )
        dialog = self.find_dialog(ICADialog)

        # Compute the components
        dialog.ui.doubleSpinBoxNComponents.setValue(0.3)
        dialog.ui.pushButtonCompute.click()

        # Select first item
        dialog.ui.listWidgetNotRemoved.setCurrentRow(0)

        # Try all the plots
        dialog.ui.pushButtonPlotTopographies.click()
        dialog.ui.pushButtonPlotSources.click()
        dialog.ui.pushButtonPlotProperties.click()

        # Transfer to be removed
        dialog.ui.pushButtonTransfer.click()

        # Try plotting changes
        dialog.ui.pushButtonPlotChanges.click()

        # Finish
        dialog.accept()



"""Defines TFR class, wraps mne.AverageTFR objects."""

import os
import re

import mne

from meggie.utilities.datatype import Datatype


class TFR(Datatype):
    """A wrapper for mne.AverageTFR objects.

    Parameters
    ----------
    name : str
        Name of the tfr, used in the UI lists and in the .exp file.
    directory : str
        Absolute path to the data folder, usually workspace/experiment/subject/tfrs.
    params : dict
        Contains additional information about the tfr.
    content : dict of mne.AverageTFR, optional
        Stores mne.AverageTFR objects as values, conditions as keys. If not provided,
        is assumed to be saved to file system earlier.
    """

    def __init__(self, name, directory, params, content=None):
        self._name = name
        self._params = params
        self._directory = directory

        # ensure comments are set to match the keys / conditions
        self._content = content
        if self._content:
            for key in self._content.keys():
                self._content[key].comment = key

    def _clean_condition_name(self, name):
        return str(name).replace(" ", "_")

    def save_content(self):
        """Saves the mne.AverageTFR to h5 files in the tfr directory."""
        try:
            for tfr_name, mne_tfr in self._content.items():

                # fail quickly for very old-style tfr's
                assert tfr_name

                fname = os.path.join(
                    self._directory,
                    self._name + "-" + self._clean_condition_name(tfr_name) + "-tfr.h5",
                )

                mne_tfr.save(fname, overwrite=True)
        except Exception:
            raise Exception(
                "Writing TFRs failed. Please ensure that the "
                "entire experiment folder has write permissions."
            )

    def delete_content(self):
        """Deletes the correct h5 files in the tfr directory"""

        template = self._name + "-" + r"([a-zA-Z1-9_-]+)\-tfr\.h5"
        for fname in os.listdir(self._directory):
            match = re.match(template, fname)
            if match:
                try:
                    key = str(match.group(1))
                except Exception:
                    continue

                # if proper condition parameters set,
                # check if the key is in there
                if "conditions" in self._params:
                    if key not in [
                        self._clean_condition_name(elem)
                        for elem in self._params["conditions"]
                    ]:
                        continue

                os.remove(os.path.join(self._directory, fname))

    def _load_content(self):
        """Handle the loading of the content."""
        self._content = {}
        template = self._name + "-" + r"([a-zA-Z1-9_-]+)\-tfr\.h5"
        for fname in os.listdir(self._directory):
            match = re.match(template, fname)
            if match:
                try:
                    fname_key = str(match.group(1))
                except Exception:
                    raise Exception("Unknown file name format.")

                condition_key = fname_key

                # skip this for old-style tfrs
                if "conditions" in self._params:
                    for condition_name in self._params["conditions"]:
                        if fname_key == self._clean_condition_name(condition_name):
                            condition_key = condition_name
                            break
                    else:
                        continue

                path = os.path.join(self._directory, fname)
                self._content[condition_key] = mne.time_frequency.read_tfrs(path)

    @property
    def content(self):
        """Returns the actual mne.AverageTFR objects either
        from cache or from the file system.
        """
        if self._content is None:
            self._load_content()
        return self._content

    @property
    def params(self):
        """Returns additional information stored."""
        return self._params

    @property
    def data(self):
        """Returns the actual numerical data of TFRs."""
        data = {}
        for key in self.content.keys():
            data[key] = self.content[key].data
        return data

    @property
    def name(self):
        """Returns the name of the tfr."""
        return self._name

    @property
    def decim(self):
        """Returns the decimation factor used in the creation."""
        return self._params.get("decim")

    @property
    def n_cycles(self):
        """ "Returns the number of cycles in used in the creation."""
        return self._params.get("n_cycles")

    @property
    def ch_names(self):
        """Returns the channel names, must read the data first."""
        return list(self.content.values())[0].info["ch_names"]

    @property
    def times(self):
        """Returns the array of times, must read the data first."""
        return list(self.content.values())[0].times

    @property
    def freqs(self):
        """Returns the array of freqs, must read the data first."""
        return list(self.content.values())[0].freqs

    @property
    def info(self):
        """Returns the info structure, must read the data first."""
        return list(self.content.values())[0].info

    @property
    def evoked_subtracted(self):
        """Returns whether evoked was subtracted in the creation."""
        return self._params.get("evoked_subtracted")



import numpy as np
import json
from meggie.utilities.serialization import serialize_dict, deserialize_dict


def test_serialization_deserialization():
    # A real-like test dict containing params with numpy arrays
    message_dict = {
        "params": {
            # np.array
            "freqs": np.array(
                [
                    5.0,
                    5.5,
                    6.0,
                    6.5,
                    7.0,
                    7.5,
                    8.0,
                    8.5,
                    9.0,
                    9.5,
                ]
            ),
            # int
            "decim": 1,
            # bool
            "subtract_evoked": False,
            # str
            "name": "TFR",
            # list
            "conditions": ["Somato"],
            # dict
            "dict_example": {"cats": "meow", "dogs": "woof"},
        }
    }

    # Serialize the dictionary
    serialized_message = serialize_dict(message_dict)
    serialized_json = json.dumps(serialized_message)

    # Simulate writing to and reading from a log file
    deserialized_json = json.loads(serialized_json)
    deserialized_message = deserialize_dict(deserialized_json)

    # Assertions to verify correctness
    for key in message_dict["params"]:
        original_value = message_dict["params"][key]
        deserialized_value = deserialized_message["params"][key]

        if isinstance(original_value, np.ndarray):
            assert isinstance(deserialized_value, np.ndarray)
            assert np.array_equal(original_value, deserialized_value)
            assert original_value.dtype == deserialized_value.dtype
            assert original_value.shape == deserialized_value.shape
        elif isinstance(original_value, list):
            assert isinstance(deserialized_value, list)
            assert original_value == deserialized_value
        else:
            assert type(original_value) is type(deserialized_value)
            assert original_value == deserialized_value



"""Contains controlling logic for the filter."""


def filter_data(subject, params, preview=False):
    """
    Filters the data array in place according to parameters in paramDict.
    Depending on the parameters, the filter is one or more of
    lowpass, highpass and bandstop (notch) filter.
    """

    raw = subject.get_raw()

    if preview:
        raw = raw.copy()

    hfreq = params["low_cutoff_freq"] if params["lowpass"] else None
    lfreq = params["high_cutoff_freq"] if params["highpass"] else None
    length = params["length"]
    trans_bw = params["trans_bw"]

    raw.filter(
        l_freq=lfreq,
        h_freq=hfreq,
        filter_length=length,
        l_trans_bandwidth=trans_bw,
        h_trans_bandwidth=trans_bw,
        method="fft",
        fir_design="firwin",
    )

    freqs = list()
    if params["bandstop1"]:
        freqs.append(params["bandstop1_freq"])
    if params["bandstop2"]:
        freqs.append(params["bandstop2_freq"])
    if len(freqs) > 0:
        length = params["bandstop_length"]
        trans_bw = params["bandstop_transbw"]

        raw.notch_filter(
            freqs,
            picks=None,
            filter_length=length,
            notch_widths=params["bandstop_bw"],
            trans_bandwidth=trans_bw,
        )

    if not preview:
        subject.save()

    return raw



from meggie.utilities.testing import BaseTestAction
from meggie.actions.evoked_delete_from_all import DeleteEvokedFromAll


class TestEvokedDeleteFromAll(BaseTestAction):
    def test_evoked_delete_from_all(self):

        data = {"outputs": {"evoked": ["Evoked"]}}

        self.run_action(
            action_name="evoked_delete_from_all",
            handler=DeleteEvokedFromAll,
            data=data,
        )



from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_filter import Filter
from meggie.actions.raw_filter.dialogs.filterDialogMain import (
    FilterDialog,
)


class TestRawFilter(BaseTestAction):
    def test_raw_filter(self):

        self.run_action(
            action_name="raw_filter",
            handler=Filter,
            patch_paths=["meggie.actions.raw_filter.dialogs.filterDialogMain"],
        )
        dialog = self.find_dialog(FilterDialog)
        dialog.accept()



"""Contains a class for logic of the group selection dialog."""

from PyQt5 import QtWidgets
from PyQt5 import QtCore

from meggie.utilities.dialogs.groupSelectionDialogUi import Ui_groupSelectionDialog


class GroupSelectionDialog(QtWidgets.QDialog):
    """Contains the logic for group selection dialog."""

    def __init__(self, experiment, parent, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_groupSelectionDialog()
        self.ui.setupUi(self)

        self.handler = handler

        subjects = experiment.subjects.keys()

        # fill the dialog with subjects
        for idx, subject_name in enumerate(subjects):
            self._add_item(idx, subject_name)

        self.subjects = subjects

    def _add_item(self, idx, name):
        setattr(self.ui, "horizontalLayoutGroup_" + str(idx), QtWidgets.QHBoxLayout())
        getattr(self.ui, "horizontalLayoutGroup_" + str(idx)).setObjectName(
            "horizontalLayoutGroup_" + str(idx)
        )

        setattr(
            self.ui,
            "checkBoxGroup_" + str(idx),
            QtWidgets.QCheckBox(self.ui.groupBoxGroups),
        )
        getattr(self.ui, "checkBoxGroup_" + str(idx)).setObjectName(
            "checkBoxGroup_" + str(idx)
        )
        getattr(self.ui, "checkBoxGroup_" + str(idx)).setText("")
        getattr(self.ui, "checkBoxGroup_" + str(idx)).setMaximumSize(20, 20)
        getattr(self.ui, "checkBoxGroup_" + str(idx)).setCheckState(
            QtCore.Qt.CheckState.Checked
        )
        getattr(self.ui, "horizontalLayoutGroup_" + str(idx)).addWidget(
            getattr(self.ui, "checkBoxGroup_" + str(idx))
        )

        setattr(
            self.ui, "labelGroup_" + str(idx), QtWidgets.QLabel(self.ui.groupBoxGroups)
        )
        getattr(self.ui, "labelGroup_" + str(idx)).setObjectName(
            "labelGroup_" + str(idx)
        )
        getattr(self.ui, "labelGroup_" + str(idx)).setText(name)
        getattr(self.ui, "horizontalLayoutGroup_" + str(idx)).addWidget(
            getattr(self.ui, "labelGroup_" + str(idx))
        )

        setattr(
            self.ui,
            "spinBoxGroup_" + str(idx),
            QtWidgets.QSpinBox(self.ui.groupBoxGroups),
        )
        getattr(self.ui, "spinBoxGroup_" + str(idx)).setMinimum(1)
        getattr(self.ui, "spinBoxGroup_" + str(idx)).setMaximumSize(40, 1000)
        getattr(self.ui, "spinBoxGroup_" + str(idx)).setObjectName(
            "spinBoxGroup_" + str(idx)
        )
        getattr(self.ui, "horizontalLayoutGroup_" + str(idx)).addWidget(
            getattr(self.ui, "spinBoxGroup_" + str(idx))
        )

        self.ui.gridLayout.addLayout(
            getattr(self.ui, "horizontalLayoutGroup_" + str(idx)), idx, 2, 1, 1
        )

    def accept(self):
        groups = {}
        for idx, subject in enumerate(self.subjects):
            selected = getattr(self.ui, "checkBoxGroup_" + str(idx)).checkState()
            if selected != QtCore.Qt.CheckState.Checked:
                continue
            group_id = getattr(self.ui, "spinBoxGroup_" + str(idx)).value()
            if group_id in groups:
                groups[group_id].append(subject)
            else:
                groups[group_id] = [subject]

        self.handler(groups)

        self.close()



"""Contains helpers to serialize and deserialize action params without losing types."""

import numpy as np


def serialize_leaf(value):
    """Serialize a leaf node in the dictionary."""

    if isinstance(value, np.ndarray):
        return {
            "is_value": True,
            "value": value.tolist(),
            "type": "numpy.ndarray",
            "dtype": str(value.dtype),
            "shape": value.shape,
        }
    elif isinstance(value, (int, float, bool, str, list)):
        return {"is_value": True, "value": value, "type": type(value).__name__}
    else:
        # Handle additional types as needed
        return {"is_value": True, "value": value, "type": type(value).__name__}


def serialize_dict(dct):
    """Recursively serialize a dictionary, converting non-dict leaves to structured dicts."""

    if isinstance(dct, dict):
        return {k: serialize_dict(v) for k, v in dct.items()}
    else:
        return serialize_leaf(dct)


def deserialize_leaf(leaf):
    """Deserialize a serialized leaf node back to its original type."""

    if "is_value" in leaf and leaf["is_value"]:
        value_type = leaf["type"]
        if value_type == "numpy.ndarray":
            return np.array(leaf["value"], dtype=leaf["dtype"]).reshape(leaf["shape"])
        elif value_type == "int":
            return int(leaf["value"])
        elif value_type == "float":
            return float(leaf["value"])
        elif value_type == "bool":
            return bool(leaf["value"])
        elif value_type == "str":
            return str(leaf["value"])
        elif value_type == "list":
            return list(leaf["value"])
        else:
            # Handle additional types as needed
            return leaf["value"]
    return leaf


def deserialize_dict(dct):
    """Recursively deserialize a dictionary, converting structured dicts back to original types."""

    if isinstance(dct, dict):
        if "is_value" in dct and dct["is_value"]:
            return deserialize_leaf(dct)
        else:
            return {k: deserialize_dict(v) for k, v in dct.items()}
    else:
        # Be brave and return the value as-is for backwards compatibility
        return dct



"""Contains functions that are used in the dynamic location and creation
of tabs and datatypes. Can be used both internally and externally.
"""

import os
import json
import importlib
import logging

from meggie.utilities.filemanager import get_resource_filename, get_package_keys

from PyQt5 import QtWidgets

from meggie.utilities.uid import generate_uid
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox
from meggie.utilities.threading import threaded
from meggie.utilities.serialization import serialize_dict


def find_all_plugins():
    """Looks for plugins (installed packages with name meggie_*), that
    can contain tabs or datatypes"""
    plugins = []
    package_keys = get_package_keys()
    for key in package_keys:
        try:
            if key.startswith("meggie_"):
                # check that there exists configuration.json
                if not os.path.exists(get_resource_filename(key, "configuration.json")):
                    continue
                plugins.append(key)
        except Exception:
            logging.getLogger("ui_logger").exception("")
    return plugins


def find_all_sources():
    """Returns all packages where to look for actions / datatypes."""
    return ["meggie"] + find_all_plugins()


def find_all_package_specs():
    """Returns all package specifications found."""
    package_specs = {}

    sources = find_all_sources()
    for source in sources:
        config_path = get_resource_filename(source, "configuration.json")
        if not os.path.exists(config_path):
            continue
        with open(config_path, "r") as f:
            config = json.load(f)
            if config and config.get("id"):
                package_specs[source] = config

    # add possibly missing fields
    for package_spec in package_specs.values():
        if "name" not in package_spec:
            package_spec["name"] = package_spec["id"]

        if "description" not in package_spec:
            package_spec["description"] = package_spec.get("name", "")

        if "author" not in package_spec:
            package_spec["author"] = ""

    return package_specs


def find_all_datatype_specs():
    """Returns all datatype specifications found."""
    datatype_specs = {}
    found_keys = []

    sources = find_all_sources()
    for source in sources:
        datatype_path = get_resource_filename(source, "datatypes")
        if not os.path.exists(datatype_path):
            continue
        for package in os.listdir(datatype_path):
            config_path = os.path.join(datatype_path, package, "configuration.json")
            if os.path.exists(config_path):
                with open(config_path, "r") as f:
                    config = json.load(f)
                    if config:
                        if config["id"] in found_keys:
                            raise Exception(
                                "Datatype with the same id found "
                                + "in multiple packages"
                            )
                        else:
                            found_keys.append(config["id"])

                        datatype_specs[config["id"]] = source, package, config

    # add possibly missing fields
    for source, package, datatype_spec in datatype_specs.values():
        if "name" not in datatype_spec:
            datatype_spec["name"] = datatype_spec["id"]

    return datatype_specs


def find_all_action_specs():
    """Returns all action specifications found."""
    action_specs = {}
    found_keys = []

    sources = find_all_sources()
    for source in sources:
        action_path = get_resource_filename(source, "actions")
        if not os.path.exists(action_path):
            continue
        for package in os.listdir(action_path):
            config_path = os.path.join(action_path, package, "configuration.json")
            if os.path.exists(config_path):
                with open(config_path, "r") as f:
                    config = json.load(f)
                    if config:
                        if config["id"] in found_keys:
                            raise Exception(
                                "Action with the same id found "
                                + "in multiple packages"
                            )
                        else:
                            found_keys.append(config["id"])

                        action_specs[config["id"]] = source, package, config

    # add possibly missing fields
    for source, package, action_spec in action_specs.values():

        if "name" not in action_spec:
            action_spec["name"] = action_spec["id"]

    return action_specs


def construct_tab(tab_spec, action_specs, datatype_specs, has_raw, parent):
    """Constructs analysis tab dynamically. Returns a QWidget
    that can be used within a QTabDialog of the main window.

    Parameters
    ----------
    tab_spec : dict
        The specification of the tab read to a dict
    action_specs : dict
        Specifications of actions
    datatype_specs : dict
        Specifications of datatypes
    parent : instance of main window
        The main window, is passed to the handlers in the ui.py.

    Returns
    -------
    instance of QWidget
        The constructed tab than can be added to
        main window.
    """

    class DynamicTab(QtWidgets.QWidget):
        """Class defining a tab.

        Parameters
        ----------
        parent : instance of main window
            The main window, is passed to the handlers in actions.
        """

        def __init__(self, parent):
            QtWidgets.QWidget.__init__(self)
            self.parent = parent
            self.tab_spec = tab_spec
            self.action_specs = action_specs
            self.datatype_specs = datatype_specs

            # first create basic layout
            self.gridLayoutContainer = QtWidgets.QGridLayout(self)

            self.gridLayoutRoot = QtWidgets.QGridLayout()
            self.gridLayoutContainer.addLayout(self.gridLayoutRoot, 0, 0, 1, 1)

            if self.tab_spec["outputs"]:
                self.groupBoxOutputs = QtWidgets.QGroupBox(self)
                self.groupBoxOutputs.setTitle("Outputs")
                self.gridLayoutOutputs = QtWidgets.QGridLayout(self.groupBoxOutputs)

            if self.tab_spec["output_actions"]:
                self.groupBoxOutputActions = QtWidgets.QGroupBox(self)
                self.groupBoxOutputActions.setTitle("")
                self.gridLayoutOutputActions = QtWidgets.QGridLayout(
                    self.groupBoxOutputActions
                )

            if self.tab_spec["input_actions"]:
                self.groupBoxInputActions = QtWidgets.QGroupBox(self)
                self.groupBoxInputActions.setTitle("")
                self.gridLayoutInputActions = QtWidgets.QGridLayout(
                    self.groupBoxInputActions
                )

            if self.tab_spec["inputs"]:
                self.groupBoxInputs = QtWidgets.QGroupBox(self)
                self.groupBoxInputs.setTitle("Inputs")
                self.gridLayoutInputs = QtWidgets.QGridLayout(self.groupBoxInputs)

            if self.tab_spec["info"]:
                self.groupBoxInfo = QtWidgets.QGroupBox(self)
                self.groupBoxInfo.setTitle("Info")
                self.gridLayoutInfo = QtWidgets.QGridLayout(self.groupBoxInfo)

            # add the (empty) input lists
            for idx, input_name in enumerate(self.tab_spec["inputs"]):

                title = input_name.capitalize()

                for source, package, datatype_spec in datatype_specs.values():
                    if datatype_spec["id"] == input_name:
                        title = datatype_spec["name"]

                groupBoxInputElement = QtWidgets.QGroupBox(self.groupBoxInputs)
                groupBoxInputElement.setTitle(title)
                gridLayoutInputElement = QtWidgets.QGridLayout(groupBoxInputElement)
                listWidgetInputElement = QtWidgets.QListWidget(groupBoxInputElement)

                listWidgetInputElement.setSelectionMode(
                    QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection
                )

                gridLayoutInputElement.addWidget(listWidgetInputElement, idx, 0, 1, 1)
                self.gridLayoutInputs.addWidget(groupBoxInputElement)

                setattr(
                    self, "groupBoxInputElement_" + str(idx + 1), groupBoxInputElement
                )
                setattr(
                    self,
                    "gridLayoutInputElement_" + str(idx + 1),
                    gridLayoutInputElement,
                )
                setattr(
                    self,
                    "listWidgetInputElement_" + str(idx + 1),
                    listWidgetInputElement,
                )

            # add the (empty) output lists
            for idx, output_name in enumerate(self.tab_spec["outputs"]):

                title = output_name.capitalize()

                for source, package, datatype_spec in datatype_specs.values():
                    if datatype_spec["id"] == output_name:
                        title = datatype_spec["name"]

                groupBoxOutputElement = QtWidgets.QGroupBox(self.groupBoxOutputs)
                groupBoxOutputElement.setTitle(title)
                gridLayoutOutputElement = QtWidgets.QGridLayout(groupBoxOutputElement)
                listWidgetOutputElement = QtWidgets.QListWidget(groupBoxOutputElement)

                listWidgetOutputElement.setSelectionMode(
                    QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection
                )

                gridLayoutOutputElement.addWidget(listWidgetOutputElement, idx, 0, 1, 1)
                self.gridLayoutOutputs.addWidget(groupBoxOutputElement)

                setattr(
                    self, "groupBoxOutputElement_" + str(idx + 1), groupBoxOutputElement
                )
                setattr(
                    self,
                    "gridLayoutOutputElement_" + str(idx + 1),
                    gridLayoutOutputElement,
                )
                setattr(
                    self,
                    "listWidgetOutputElement_" + str(idx + 1),
                    listWidgetOutputElement,
                )

            # add input action buttons
            for idx, action_name in enumerate(self.tab_spec["input_actions"]):

                action_spec = self.action_specs[action_name][2]
                title = action_spec["name"]
                description = action_spec.get("description", "")

                pushButtonInputActionElement = QtWidgets.QPushButton(
                    self.groupBoxInputActions
                )

                if "raw" in action_spec.get("tags", []) and not has_raw:
                    pushButtonInputActionElement.setEnabled(False)

                pushButtonInputActionElement.setText(title)
                pushButtonInputActionElement.setToolTip(description)
                self.gridLayoutInputActions.addWidget(
                    pushButtonInputActionElement, idx, 0, 1, 1
                )
                setattr(
                    self,
                    "pushButtonInputActionElement_" + str(idx + 1),
                    pushButtonInputActionElement,
                )

            if getattr(self, "gridLayoutInputActions", None):
                spacer = QtWidgets.QSpacerItem(
                    20,
                    10,
                    QtWidgets.QSizePolicy.Policy.Minimum,
                    QtWidgets.QSizePolicy.Policy.Expanding,
                )
                self.gridLayoutInputActions.addItem(spacer, idx + 1, 0, 1, 1)

            # add output action buttons
            for idx, action_name in enumerate(self.tab_spec["output_actions"]):

                action_spec = self.action_specs[action_name][2]
                title = action_spec["name"]
                description = action_spec.get("description", "")

                pushButtonOutputActionElement = QtWidgets.QPushButton(
                    self.groupBoxOutputActions
                )

                if "raw" in action_spec.get("tags", []) and not has_raw:
                    pushButtonOutputActionElement.setEnabled(False)

                pushButtonOutputActionElement.setText(title)
                pushButtonOutputActionElement.setToolTip(description)
                self.gridLayoutOutputActions.addWidget(
                    pushButtonOutputActionElement, idx, 0, 1, 1
                )
                setattr(
                    self,
                    "pushButtonOutputActionElement_" + str(idx + 1),
                    pushButtonOutputActionElement,
                )

            if getattr(self, "gridLayoutOutputActions", None):
                spacer = QtWidgets.QSpacerItem(
                    20,
                    10,
                    QtWidgets.QSizePolicy.Policy.Minimum,
                    QtWidgets.QSizePolicy.Policy.Expanding,
                )
                self.gridLayoutOutputActions.addItem(spacer, idx + 1, 0, 1, 1)

            # add info text boxes
            for idx, info_name in enumerate(self.tab_spec["info"]):

                action_spec = self.action_specs[info_name][2]
                title = action_spec["name"]

                groupBoxInfoElement = QtWidgets.QGroupBox(self.groupBoxInfo)
                groupBoxInfoElement.setTitle(title)
                gridLayoutInfoElement = QtWidgets.QGridLayout(groupBoxInfoElement)
                plainTextEditInfoElement = QtWidgets.QTextBrowser(groupBoxInfoElement)

                gridLayoutInfoElement.addWidget(plainTextEditInfoElement, idx, 0, 1, 1)
                self.gridLayoutInfo.addWidget(groupBoxInfoElement)

                setattr(
                    self, "groupBoxInfoElement_" + str(idx + 1), groupBoxInfoElement
                )
                setattr(
                    self, "gridLayoutInfoElement_" + str(idx + 1), gridLayoutInfoElement
                )
                setattr(
                    self,
                    "plainTextEditInfoElement_" + str(idx + 1),
                    plainTextEditInfoElement,
                )

            # lay out inputs, outputs, actions and info elements
            # in a nice way to a grid
            if self.tab_spec["inputs"] and not self.tab_spec["input_actions"]:
                self.gridLayoutRoot.addWidget(self.groupBoxInputs, 0, 0, 2, 1)
            elif self.tab_spec["inputs"]:
                self.gridLayoutRoot.addWidget(self.groupBoxInputs, 0, 0, 1, 1)

            if self.tab_spec["outputs"] and not self.tab_spec["output_actions"]:
                self.gridLayoutRoot.addWidget(self.groupBoxOutputs, 0, 1, 2, 1)
            elif self.tab_spec["outputs"]:
                self.gridLayoutRoot.addWidget(self.groupBoxOutputs, 0, 1, 1, 1)

            if self.tab_spec["input_actions"] and not self.tab_spec["inputs"]:
                self.gridLayoutRoot.addWidget(self.groupBoxInputActions, 0, 0, 2, 1)
            elif self.tab_spec["input_actions"]:
                self.gridLayoutRoot.addWidget(self.groupBoxInputActions, 1, 0, 1, 1)

            if self.tab_spec["output_actions"] and not self.tab_spec["outputs"]:
                self.gridLayoutRoot.addWidget(self.groupBoxOutputActions, 0, 1, 2, 1)
            elif self.tab_spec["output_actions"]:
                self.gridLayoutRoot.addWidget(self.groupBoxOutputActions, 1, 1, 1, 1)

            if self.tab_spec["info"]:
                self.gridLayoutRoot.addWidget(self.groupBoxInfo, 0, 2, 2, 1)

            # add spacers to bottom and right to keep the window concise
            spacerItemVertical = QtWidgets.QSpacerItem(
                20,
                10,
                QtWidgets.QSizePolicy.Policy.Minimum,
                QtWidgets.QSizePolicy.Policy.Expanding,
            )
            self.gridLayoutContainer.addItem(spacerItemVertical, 1, 0, 1, 1)
            spacerItemHorizontal = QtWidgets.QSpacerItem(
                10,
                20,
                QtWidgets.QSizePolicy.Policy.Expanding,
                QtWidgets.QSizePolicy.Policy.Minimum,
            )
            self.gridLayoutContainer.addItem(spacerItemHorizontal, 0, 1, 1, 1)

            # add handlers for list selection changed -> info updates
            def connect_to_handler(list_element, info_element, info_name):

                source, package, action_spec = self.action_specs[info_name]
                module = importlib.import_module(".".join([source, "actions", package]))
                entry = action_spec["entry"]

                @threaded
                def handler(*args):
                    return getattr(module, entry)(*args).run()

                def handler_wrapper():
                    experiment = self.parent.experiment
                    if not experiment:
                        return
                    subject = experiment.active_subject
                    if not subject:
                        return

                    if "raw" in action_spec.get("tags", []) and not has_raw:
                        return

                    data = self._get_data()

                    try:
                        info_content = handler(
                            experiment,
                            data,
                            parent,
                            action_spec,
                            do_meanwhile=parent.update_ui,
                        )
                        info_element.setPlainText(info_content)
                    except Exception as exc:
                        exc_messagebox(self, exc)

                list_element.itemSelectionChanged.connect(handler_wrapper)

            for idx, info_name in enumerate(self.tab_spec["info"]):
                info_element = getattr(self, "plainTextEditInfoElement_" + str(idx + 1))

                for idx, input_name in enumerate(self.tab_spec["inputs"]):

                    input_element = getattr(
                        self, "listWidgetInputElement_" + str(idx + 1)
                    )
                    connect_to_handler(input_element, info_element, info_name)

                for idx, output_name in enumerate(self.tab_spec["outputs"]):

                    output_element = getattr(
                        self, "listWidgetOutputElement_" + str(idx + 1)
                    )
                    connect_to_handler(output_element, info_element, info_name)

            # add button handlers
            def connect_to_handler(button, name):

                source, package, action_spec = self.action_specs[name]
                module = importlib.import_module(".".join([source, "actions", package]))
                entry = action_spec["entry"]

                handler = getattr(module, entry)

                def handler_wrapper(checked):
                    experiment = self.parent.experiment
                    if not experiment:
                        messagebox(self, "You need to open an experiment first.")
                        return

                    subject = experiment.active_subject
                    if not subject:
                        messagebox(self, "You need to activate a subject first.")
                        return

                    data = self._get_data()

                    try:
                        handler(experiment, data, parent, action_spec).run()
                    except Exception as exc:
                        exc_messagebox(self, exc)

                button.clicked.connect(handler_wrapper)

            for idx, action_name in enumerate(self.tab_spec["input_actions"]):
                action_element = getattr(
                    self, "pushButtonInputActionElement_" + str(idx + 1)
                )
                connect_to_handler(action_element, action_name)

            for idx, action_name in enumerate(self.tab_spec["output_actions"]):
                action_element = getattr(
                    self, "pushButtonOutputActionElement_" + str(idx + 1)
                )
                connect_to_handler(action_element, action_name)

        def _get_data(self):
            """Returns data from input and output lists."""
            data = {"inputs": {}, "outputs": {}}

            data["tab_id"] = self.tab_spec["id"]

            for idx, name in enumerate(self.tab_spec["inputs"]):
                ui_element = getattr(self, "listWidgetInputElement_" + str(idx + 1))
                try:
                    selected_items = [
                        item.text() for item in ui_element.selectedItems()
                    ]
                except BaseException:
                    continue

                data["inputs"][name] = selected_items

            for idx, name in enumerate(self.tab_spec["outputs"]):
                ui_element = getattr(self, "listWidgetOutputElement_" + str(idx + 1))
                try:
                    selected_items = [
                        item.text() for item in ui_element.selectedItems()
                    ]
                except BaseException:
                    continue

                data["outputs"][name] = selected_items

            return data

        def initialize_ui(self):
            """Updates (empties and refills) ui contents when called."""
            experiment = self.parent.experiment
            if not experiment:
                return

            subject = experiment.active_subject

            # fill input lists with contents
            for idx, input_name in enumerate(self.tab_spec["inputs"]):
                ui_element = getattr(self, "listWidgetInputElement_" + str(idx + 1))

                ui_element.clear()

                if not subject:
                    continue

                data = getattr(subject, input_name, None)
                if not data:
                    continue

                for key in sorted(list(data.keys())):
                    ui_element.addItem(key)

            # fill output lists with contents
            for idx, output_name in enumerate(self.tab_spec["outputs"]):
                ui_element = getattr(self, "listWidgetOutputElement_" + str(idx + 1))

                ui_element.clear()

                if not subject:
                    continue

                data = getattr(subject, output_name, None)
                if not data:
                    continue

                for key in sorted(list(data.keys())):
                    ui_element.addItem(key)

            # allow to fill info element already here.
            # there are also handlers to update info element
            # on list selection changes
            for idx, info_name in enumerate(self.tab_spec["info"]):

                ui_element = getattr(self, "plainTextEditInfoElement_" + str(idx + 1))

                source, package, action_spec = self.action_specs[info_name]
                module = importlib.import_module(".".join([source, "actions", package]))
                entry = action_spec["entry"]

                if not subject:
                    continue

                if "raw" in action_spec.get("tags", []) and not has_raw:
                    return

                @threaded
                def handler(*args):
                    return getattr(module, entry)(*args).run()

                try:
                    info_content = handler(
                        experiment,
                        None,
                        self.parent,
                        action_spec,
                        do_meanwhile=parent.update_ui,
                    )
                    ui_element.setPlainText(info_content)
                except Exception as exc:
                    exc_messagebox(self, exc)

        @property
        def name(self):
            """Returns name of the tab."""
            return self.tab_spec["name"]

    DynamicTab.__name__ = "MainWindowTab" + tab_spec["id"].capitalize()

    tab = DynamicTab(parent)

    return tab


def construct_tabs(selected_pipeline, window, prefs, include_eeg, has_raw):
    """Constructs as set of tabs based on specifications and the
    selected pipeline.

    Parameters
    ----------
    selected_pipeline : str
        ID of the selected pipeline
    window : instance of main window
        The main window.
    prefs : Instance of PreferencesHandler
        Stores e.g. active plugins.
    include_eeg : bool
        Whether to add EEG-related actions

    Returns
    -------
    list of QWidget
        Contains the constructed tabs relevant to the pipeline
    """

    active_plugins = prefs.active_plugins

    action_specs = find_all_action_specs()
    datatype_specs = find_all_datatype_specs()
    package_specs = find_all_package_specs()

    tabs = []
    pipelines = []
    for source, package_spec in package_specs.items():
        if source not in active_plugins and source != "meggie":
            continue
        if "tabs" in package_spec:
            tabs.extend(package_spec["tabs"])
        if "pipelines" in package_spec:
            pipelines.extend(package_spec["pipelines"])

    # add possibly missing fields
    for tab in tabs:
        if "id" not in tab:
            raise Exception("Every tab specification must have id.")

        if "name" not in tab:
            tab["name"] = tab["id"]

        if "inputs" not in tab:
            tab["inputs"] = []

        if "outputs" not in tab:
            tab["outputs"] = []

        if "input_actions" not in tab:
            tab["input_actions"] = []

        if "output_actions" not in tab:
            tab["output_actions"] = []

        if "info" not in tab:
            tab["info"] = []

    for pipeline in pipelines:
        if "id" not in pipeline:
            raise Exception("Every pipeline must have id.")

        if "name" not in pipeline:
            pipeline["name"] = pipeline["id"]

    pipeline_specs = []
    for pipeline in pipelines:
        # there might be multiple configuration for the same pipeline from addons
        if pipeline["id"] == selected_pipeline:
            pipeline_specs.append(pipeline)

    if not pipeline_specs:
        # Use classic
        pipeline_specs = [{"id": "classic", "name": "Include everything"}]

    # merges tab specification from others to first and
    # filters to tabs specified by the pipeline
    combined_tabs = []
    for tab_spec in tabs:

        # Include only tabs relevant to the pipeline
        if all(
            [
                spec["id"] != "classic"
                and tab_spec["id"] not in spec.get("include_tabs", [])
                for spec in pipeline_specs
            ]
        ):
            continue

        # if a completely new tab, initialize it
        if tab_spec["id"] not in [tab["id"] for tab in combined_tabs]:

            new_tab = {}
            new_tab["id"] = tab_spec["id"]
            new_tab["name"] = tab_spec["name"]
            new_tab["inputs"] = []
            new_tab["outputs"] = []
            new_tab["input_actions"] = []
            new_tab["output_actions"] = []
            new_tab["info"] = []

            idx = len(combined_tabs)
            combined_tabs.append(new_tab)
        # otherwise find idx of the tab in the first specification
        else:
            idx = [tab["id"] for tab in combined_tabs].index(tab_spec["id"])

        # Add missing inputs
        for input_spec in tab_spec["inputs"]:
            if input_spec not in combined_tabs[idx]["inputs"]:
                combined_tabs[idx]["inputs"].append(input_spec)

        # Add missing outputs
        for output_spec in tab_spec["outputs"]:
            if output_spec not in combined_tabs[idx]["outputs"]:
                combined_tabs[idx]["outputs"].append(output_spec)

        # add missing input actions
        for input_spec in tab_spec["input_actions"]:

            action_spec = action_specs.get(input_spec)
            if not action_spec:
                raise Exception("Cannot read action " + input_spec + ".")

            if not include_eeg and "eeg" in action_spec[2].get("tags", []):
                continue

            if input_spec not in combined_tabs[idx]["input_actions"]:
                combined_tabs[idx]["input_actions"].append(input_spec)

        # add missing output actions
        for output_spec in tab_spec["output_actions"]:

            action_spec = action_specs.get(output_spec)
            if not action_spec:
                raise Exception("Cannot read action " + output_spec + ".")

            if not include_eeg and "eeg" in action_spec[2].get("tags", []):
                continue

            if output_spec not in combined_tabs[idx]["output_actions"]:
                combined_tabs[idx]["output_actions"].append(output_spec)

        # add missing info elements
        for info_spec in tab_spec["info"]:

            action_spec = action_specs.get(info_spec)
            if not action_spec:
                raise Exception("Cannot read info item " + info_spec + ".")

            if info_spec not in combined_tabs[idx]["info"]:
                combined_tabs[idx]["info"].append(info_spec)

    # If everything is fine, construct each of the tabs
    tabs = []
    for tab_spec in combined_tabs:
        tabs.append(
            construct_tab(tab_spec, action_specs, datatype_specs, has_raw, window)
        )

    return tabs


def subject_action(inner_function):
    """Decorator for automatically logging actions for each subject
    separately. Should be used within Action-instances.
    """

    def outer_function(self, subject, params, *args, **kwargs):

        # ensure that we have enough type information to construct
        # params on the other side
        serialized_params = serialize_dict(params)

        message_dict = {
            "uid": self.uid,
            "type": "SUBJECT_START",
            "subject_uid": subject.uid,
            "params": serialized_params,
        }
        logging.getLogger("action_logger").info(message_dict)

        res = inner_function(self, subject, params, *args, **kwargs)

        message_dict = {
            "uid": self.uid,
            "type": "SUBJECT_END",
            "subject_uid": subject.uid,
        }
        logging.getLogger("action_logger").info(message_dict)

        message = '"{0}" finished successfully for subject {1}.'.format(
            self.action_spec["name"], subject.name
        )
        logging.getLogger("ui_logger").info(message)

        if self.subject_action_callback:
            self.subject_action_callback(self.action_spec["id"], subject.name)

        return res

    outer_function._is_subject_action = True

    return outer_function


class Action:
    """All actions should inherit this, so that automatic logging
    works.
    """

    logged = True

    def __init__(
        self, experiment, data, window, action_spec, subject_action_callback=None
    ):
        """ """
        self.experiment = experiment
        self.data = data
        self.window = window
        self.action_spec = action_spec
        self.subject_action_callback = subject_action_callback

        # generate short temporary uid
        self.uid = generate_uid()

        if self.logged:
            # and log action with it
            message_dict = {
                "version": 1,
                "uid": self.uid,
                "data": self.data,
                "type": "ACTION_START",
                "id": action_spec["id"],
                "desc": self.data["tab_id"],
            }
            logging.getLogger("action_logger").info(message_dict)

    def run(self, params={}):
        """Called when action is initiated"""
        messagebox("Running action " + self.action_spec["id"])


class InfoAction(Action):
    """This should be inherited instead of Action
    when filling info boxes."""

    logged = False

    def run(self, params={}):
        """Called when the info box is filled."""
        return "Returned content."



"""Contains controlling logic for tfr create"""

from meggie.utilities.threading import threaded
from meggie.utilities.validators import assert_arrays_same

from meggie.datatypes.tfr.tfr import TFR


@threaded
def create_tfr(
    subject, tfr_name, epochs_names, freqs, decim, n_cycles, subtract_evoked
):
    """Handles tfr item creation."""

    time_arrays = []
    for name in epochs_names:
        collection = subject.epochs.get(name)
        if collection:
            time_arrays.append(collection.content.times)
    assert_arrays_same(time_arrays)

    tfrs = {}
    for epoch_name in epochs_names:
        epochs = subject.epochs[epoch_name].content
        if subtract_evoked:
            epochs = epochs.copy().subtract_evoked()

        tfr = epochs.compute_tfr(
            "morlet",
            freqs,
            n_cycles=n_cycles,
            decim=decim,
            average=True,
            return_itc=False,
        )
        tfrs[epoch_name] = tfr

    # convert list-like to list
    if hasattr(n_cycles, "__len__"):
        n_cycles = list(n_cycles)

    params = {
        "decim": decim,
        "n_cycles": n_cycles,
        "evoked_subtracted": subtract_evoked,
        "conditions": epochs_names,
    }

    meggie_tfr = TFR(tfr_name, subject.tfr_directory, params, tfrs)

    meggie_tfr.save_content()
    subject.add(meggie_tfr, "tfr")



"""Contains controlling logic for the tfr implementation"""

import numpy as np
import matplotlib.pyplot as plt
import mne

from meggie.utilities.plotting import create_channel_average_plot
from meggie.utilities.channels import average_to_channel_groups
from meggie.utilities.channels import ensure_montage
from meggie.utilities.plotting import set_figure_title


def plot_tfr_averages(
    subject,
    tfr_name,
    tfr_condition,
    blmode,
    blstart,
    blend,
    tmin,
    tmax,
    fmin,
    fmax,
    channel_groups,
):
    """Plots tfr averages."""

    meggie_tfr = subject.tfr[tfr_name]

    if blmode:
        bline = (blstart, blend)
        mode = blmode
    else:
        bline = None
        mode = None

    tfr = meggie_tfr.content.get(tfr_condition)

    ch_names = meggie_tfr.ch_names

    sfreq = meggie_tfr.info["sfreq"]
    times = meggie_tfr.times
    freqs = meggie_tfr.freqs

    # note: baseline is corrected before channel average
    data = mne.baseline.rescale(tfr.data, times, baseline=bline, mode=mode)

    data_labels, averaged_data = average_to_channel_groups(
        data, meggie_tfr.info, ch_names, channel_groups
    )

    averages = {}
    for label_idx, label in enumerate(data_labels):
        averages[label] = averaged_data[label_idx]

    ch_types = sorted(set([label[0] for label in data_labels]))
    for ch_type in ch_types:

        ch_groups = sorted([label[1] for label in data_labels if label[0] == ch_type])

        def plot_fun(ax_idx, ax):
            ch_group = ch_groups[ax_idx]
            ax.set_title(ch_group)

            info = mne.create_info(
                ch_names=["grand_average"], sfreq=sfreq, ch_types="mag"
            )
            tfr = mne.time_frequency.tfr.AverageTFRArray(
                info, averages[(ch_type, ch_group)][np.newaxis, :], times, freqs, nave=1
            )

            # prevent interaction as no topography is involved now
            def onselect(*args, **kwargs):
                pass

            tfr._onselect = onselect

            tfr.plot(
                baseline=None,
                title="",
                fmin=fmin,
                fmax=fmax,
                tmin=tmin,
                tmax=tmax,
                axes=ax,
            )

        title = " ".join([tfr_name, tfr_condition, ch_type])
        create_channel_average_plot(len(ch_groups), plot_fun, title)


def plot_tfr_topo(
    subject,
    tfr_name,
    tfr_condition,
    blmode,
    blstart,
    blend,
    tmin,
    tmax,
    fmin,
    fmax,
    ch_type,
):
    """Plots tfr topography."""

    meggie_tfr = subject.tfr[tfr_name]

    if blmode:
        bline = (blstart, blend)
        mode = blmode
    else:
        bline = None
        mode = None

    tfr = meggie_tfr.content.get(tfr_condition)

    if ch_type == "eeg":
        dropped_names = [
            ch_name
            for ch_idx, ch_name in enumerate(tfr.info["ch_names"])
            if ch_idx not in mne.pick_types(tfr.info, eeg=True, meg=False)
        ]
    elif ch_type == "grad":
        dropped_names = [
            ch_name
            for ch_idx, ch_name in enumerate(tfr.info["ch_names"])
            if ch_idx not in mne.pick_types(tfr.info, eeg=False, meg="grad")
        ]
    elif ch_type == "mag":
        dropped_names = [
            ch_name
            for ch_idx, ch_name in enumerate(tfr.info["ch_names"])
            if ch_idx not in mne.pick_types(tfr.info, eeg=False, meg="mag")
        ]

    tfr = tfr.copy().drop_channels(dropped_names)

    # apply baseline here to get better vmin and vmax values
    tfr.apply_baseline(baseline=bline, mode=mode)

    ensure_montage(subject, tfr, ch_type)

    values = tfr.data.flatten()
    vmax = np.percentile(np.abs(values), 99)
    vmin = -vmax

    title = "{0}_{1}_{2}".format(tfr_name, tfr_condition, ch_type)
    fig = tfr.plot_topo(
        show=False,
        baseline=None,
        tmin=tmin,
        tmax=tmax,
        fmin=fmin,
        fmax=fmax,
        vmin=vmin,
        vmax=vmax,
        title="",
    )

    set_figure_title(fig, title)

    def onclick(event):
        """hacky way to change title and add colorbar after creation"""
        ax = plt.gca()
        fig = plt.gcf()

        channel = plt.getp(ax, "title")
        if not channel:
            return

        ax.set_title("")

        title = " ".join([tfr_name, channel])
        set_figure_title(fig, title.replace(" ", "_"))
        fig.suptitle(title)

        img = ax.get_images()[0]
        plt.colorbar(mappable=img, ax=ax)

        plt.show(block=False)

    fig.canvas.mpl_connect("button_press_event", onclick)
    fig.show()



"""Contains ICA action handling."""

from meggie.actions.raw_ica.dialogs.icaDialogMain import ICADialog

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class ICA(Action):
    """Shows a dialog which can be used to compute ICA decompositions
    and selecting components for removal.
    """

    random_state = 10

    def run(self, params={}):
        ica_dialog = ICADialog(
            self.window, self.experiment, self.random_state, on_apply=self.handler
        )
        ica_dialog.show()

    @subject_action
    def handler(self, subject, params):
        """ """



"""Contains implementation for delete evoked"""

from meggie.utilities.messaging import exc_messagebox

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action


class DeleteEvoked(Action):
    """Deletes evoked of selected name"""

    def run(self, params={}):

        subject = self.experiment.active_subject

        try:
            selected_name = self.data["outputs"]["evoked"][0]
        except IndexError:
            return

        try:
            self.handler(subject, {"name": selected_name})
        except Exception as exc:
            exc_messagebox(self.window, exc)

        self.window.initialize_ui()

    @subject_action
    def handler(self, subject, params):
        subject.remove(params["name"], "evoked")
        self.experiment.save_experiment_settings()



from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_montage import Montage
from meggie.actions.raw_montage.dialogs.montageDialogMain import (
    MontageDialog,
)


class TestRawMontage(BaseTestAction):

    def setup_experiment(self):
        BaseTestAction.setup_experiment(self)
        raw = self.experiment.active_subject.get_raw()

        # rename channels to match the one lucky montage that comes in mne
        raw.rename_channels(
            dict(
                [
                    (old, old.split(" ")[0] + old.split(" ")[1])
                    for old in raw.info["ch_names"]
                ]
            )
        )

    def test_raw_montage(self):

        self.run_action(
            action_name="raw_montage",
            handler=Montage,
            patch_paths=["meggie.actions.raw_montage.dialogs.montageDialogMain"],
        )
        dialog = self.find_dialog(MontageDialog)
        dialog.ui.comboBoxSelectFromList.setCurrentText("mgh60.elc")
        dialog.accept()



"""Contains classes and functions for experiments."""

import os
import json
import importlib
import shutil
import logging
from meggie.utilities.filemanager import get_distribution_version

from meggie.subject import Subject

from meggie.utilities.filemanager import open_raw
from meggie.utilities.filemanager import save_raw
from meggie.utilities.channels import get_default_channel_groups
from meggie.utilities.channels import find_montage_info
from meggie.utilities.validators import validate_name
from meggie.utilities.uid import generate_uid

from meggie.mainwindow.dynamic import find_all_datatype_specs


class Experiment:
    """A top-level class that contains structure of an experiment,
    all the subjects within, and so forth.

    Parameters
    ----------
    name : str
        Name of the experiment
    author : str
        Name of the author
    path : str
        Path to the experiment folder.
    """

    def __init__(self, name, author, path):
        self._name = name
        self._author = author
        self._path = path
        self._subjects = {}
        self._active_subject = None

        self._channel_groups = {
            "eeg": {},
            "meg": {},
        }
        self._selected_pipeline = "classic"

    @property
    def path(self):
        """Returns the path."""
        return self._path

    @path.setter
    def path(self, path):
        """Sets the path."""
        self._path = path

    @property
    def name(self):
        """Returns the name."""
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name."""
        self._name = validate_name(name, fieldname="name")

    @property
    def author(self):
        """Returns the author."""
        return self._author

    @author.setter
    def author(self, author):
        """Sets the author."""
        self._author = validate_name(author, minlength=0, fieldname="author")

    @property
    def channel_groups(self):
        """Returns channel groups for experiment. If not set,
        uses defaults."""
        channel_groups = self._channel_groups.copy()
        specs = find_all_datatype_specs()

        # if channel groups not found, look for defaults..
        if not channel_groups.get("eeg"):
            if self.active_subject:
                info = find_montage_info(self.active_subject, specs.keys(), "eeg")
                if info:
                    try:
                        channel_groups["eeg"] = get_default_channel_groups(info, "eeg")
                    except Exception:
                        pass

        if not channel_groups.get("meg"):
            if self.active_subject:
                info = find_montage_info(self.active_subject, specs.keys(), "meg")
                if info:
                    try:
                        channel_groups["meg"] = get_default_channel_groups(info, "meg")
                    except Exception:
                        pass

        return channel_groups

    @channel_groups.setter
    def channel_groups(self, channel_groups):
        """Sets the channel groups."""
        self._channel_groups = channel_groups

    @property
    def selected_pipeline(self):
        """Returns selected pipeline"""
        return self._selected_pipeline

    @selected_pipeline.setter
    def selected_pipeline(self, selected_pipeline):
        """Sets selected pipeline"""
        self._selected_pipeline = selected_pipeline

    @property
    def active_subject(self):
        """Returns the active subject."""
        return self._active_subject

    @active_subject.setter
    def active_subject(self, subject):
        """Sets the active subject."""
        self._active_subject = subject

    def add_subject(self, subject):
        """Adds subject to the experiment.

        Parameters
        ----------
        subject : meggie.subject.Subject
            A subject object.
        """
        self.subjects[subject.name] = subject

    @property
    def subjects(self):
        """Returns contained subjects."""
        return self._subjects

    def remove_subject(self, name):
        """Removes a subject.

        Removes the subject folder and its contents under experiment tree, and
        removes the subject from the experiment.

        Parameters
        ----------
        name : str
            Name of the subject to be removed.

        """

        if getattr(self, "active_subject", None) and self.active_subject.name == name:
            self.active_subject = None

        subject = self.subjects.pop(name)

        try:
            shutil.rmtree(subject.path)
        except Exception:
            raise Exception("Could not remove the contents of the subject folder.")

    def activate_subject(self, subject_name):
        """Activates a subject from the existing subjects

        Parameters
        ----------
        subject_name : str
            Name of the subject.

        Returns
        -------
        meggie.subject.Subject
            The activated subject.
        """
        # remove raw files from memory before activating new subject.
        if self.active_subject:
            self.active_subject.release_memory()

        self.active_subject = self.subjects[subject_name]

        # test validity
        if self.active_subject.has_raw:
            self.active_subject.get_raw(preload=False)

        return self.active_subject

    def create_subject(self, subject_name, raw_path):
        """Creates a subject object and copies the raw file
        inside it (by reading and saving).

        Parameters
        ----------
        subject_name : str
            Name of the new subject.
        raw_path : str | None
            Path to the data file.

        Returns
        -------
        meggie.subject.Subject
            The activated subject.

        """

        uid = generate_uid()

        # it is possible to add "placeholder" subjects with only e.g epochs data
        if raw_path:
            bname = os.path.basename(raw_path)
            stem, ext = os.path.splitext(bname)
            new_fname = stem + ".fif"

            subject = Subject(self, subject_name, new_fname, uid)
            subject.ensure_folders()

            raw = open_raw(raw_path)

            new_path = os.path.join(subject.path, new_fname)
            save_raw(raw, new_path)

        else:
            subject = Subject(self, subject_name, "", uid)

        subject.ensure_folders()

        self.add_subject(subject)

        return subject

    def save_experiment_settings(self):
        """
        Saves the experiment settings into a file in the root of
        the experiment directory structure.
        """
        subjects = []

        for subject in self.subjects.values():
            subject_dict = {
                "subject_name": subject.name,
                "raw_fname": subject.raw_fname if subject.raw_fname else "",
                "uid": subject.uid,
                "ica_applied": subject.ica_applied,
                "rereferenced": subject.rereferenced,
            }

            datatype_specs = find_all_datatype_specs()

            for source, package, datatype_spec in datatype_specs.values():
                save_key = datatype_spec["save_key"]
                datatype = datatype_spec["id"]

                for inst in getattr(subject, datatype).values():
                    datatype_dict = {"name": inst.name, "params": inst.params}

                    if save_key not in subject_dict.keys():
                        subject_dict[save_key] = []

                    subject_dict[save_key].append(datatype_dict)

            subjects.append(subject_dict)

        save_dict = {
            "subjects": subjects,
            "name": self._name,
            "author": self._author,
            "channel_groups": self._channel_groups,
            "selected_pipeline": self._selected_pipeline,
        }

        version = get_distribution_version("meggie")

        save_dict["version"] = version

        try:
            os.makedirs(self.path)
        except OSError:
            pass

        path = os.path.join(self._path, os.path.basename(self._path) + ".exp")

        # let's backup previous exp file with version number
        try:
            if os.path.exists(path):
                with open(path, "r") as f:
                    old_data = json.load(f)

                version = str(old_data["version"]).replace(".", "-")

                backup_path = os.path.join(
                    self._path,
                    os.path.basename(self._path) + "_" + version + ".exp.bak",
                )

                shutil.copy(path, backup_path)
        except Exception:
            message = (
                "Could not backup experiment file to {0}. "
                "Please check that the experiment folder "
                "has write permissions everywhere."
            )
            raise Exception(message.format(backup_path))

        # and then overwrite the current exp file
        try:
            with open(path, "w") as f:
                json.dump(save_dict, f, sort_keys=True, indent=4)
        except Exception:
            message = (
                "Could not save experiment file {0}. "
                "Please check that the experiment folder "
                "has write permissions everywhere."
            )
            raise Exception(message.format(path))


def initialize_new_experiment(name, author, prefs, set_previous_experiment=True):
    """Initializes new experiment object with given data.

    Parameters
    ----------
    name : str
        Name of the experiment.
    author : str
        Name of the author.
    prefs : meggie.mainwindow.preferences.PreferencesHandler
        A preferences object.

    Returns
    -------
    meggie.experiment.Experiment
        The new experiment.

    """
    path = os.path.join(prefs.workspace, name)
    if os.path.exists(path):
        raise Exception("Experiment with same name already exists.")

    experiment = Experiment(name, author, path)
    experiment.save_experiment_settings()

    if set_previous_experiment:
        prefs.previous_experiment_name = experiment.path
        prefs.write_preferences_to_disk()
    return experiment


def open_existing_experiment(prefs, path=None):
    """Reads and opens existing experiment.

    Parameters
    ----------
    prefs : meggie.mainwindow.preferences.PreferencesHandler
        A preferences object.
    path : str, optional
        Path to the experiment folder.

    Returns
    -------
    meggie.experiment.Experiment
        The opened experiment.
    """

    experiment_updated = False

    if path:
        exp_file = os.path.join(path, os.path.basename(path) + ".exp")
    else:
        exp_file = os.path.join(
            prefs.previous_experiment_name,
            os.path.basename(prefs.previous_experiment_name) + ".exp",
        )

    if not os.path.isfile(exp_file):
        raise ValueError("Trying to open experiment without settings file (.exp).")

    try:
        with open(exp_file, "r") as f:
            data = json.load(f)
    except ValueError:
        logging.getLogger("ui_logger").exception("")
        raise ValueError(
            "Experiment from "
            + exp_file
            + " could not be "
            + "opened. There might be a problem with "
            + "cohesion of the experiment file."
        )

    if not path:
        path = os.path.dirname(exp_file)

    experiment = Experiment(data["name"], data["author"], path)

    if "channel_groups" in data.keys() and data["channel_groups"] != "MNE":
        experiment.channel_groups = data["channel_groups"]
    else:
        experiment.channel_groups = {
            "eeg": {},
            "meg": {},
        }

    if "selected_pipeline" in data.keys():
        experiment.selected_pipeline = data["selected_pipeline"]

    if len(data["subjects"]) == 0:
        return experiment

    for subject_data in data["subjects"]:

        subject_name = subject_data["subject_name"]

        raw_fname = subject_data.get("raw_fname")

        # for backwards compatibility
        if not raw_fname:
            raw_fname = subject_data.get("working_file_name")

        uid = subject_data.get("uid")
        if not uid:
            uid = generate_uid()
            experiment_updated = True

        subject = Subject(
            experiment,
            subject_name,
            raw_fname,
            uid,
            ica_applied=subject_data.get("ica_applied", False),
            rereferenced=subject_data.get("rereferenced", False),
        )

        datatype_specs = find_all_datatype_specs()

        for source, package, datatype_spec in datatype_specs.values():
            save_key = datatype_spec["save_key"]
            entry = datatype_spec["entry"]
            datatype = datatype_spec["id"]

            for inst_data in subject_data.get(save_key, []):
                module_name, class_name = entry.split(".")
                module = importlib.import_module(
                    ".".join([source, "datatypes", package, module_name])
                )
                inst_class = getattr(module, class_name, None)

                name = inst_data.get("name")
                # backward compatibility
                if not name and "collection_name" in inst_data:
                    name = inst_data.get("collection_name")
                if not name:
                    raise Exception("No name attribute found")

                directory = getattr(subject, datatype + "_directory")
                params = inst_data.get("params", {})

                # for backwards compatibility
                if datatype == "evoked":
                    if not params.get("conditions") and "event_names" in inst_data:
                        params["conditions"] = inst_data["event_names"]
                    params["bwc_path"] = os.path.join(subject.path, "epochs/average")
                if datatype == "spectrum":
                    if (
                        not params.get("log_transformed")
                        and "log_transformed" in inst_data
                    ):
                        params["log_transformed"] = inst_data["log_transformed"]
                if datatype == "tfr":
                    if not params.get("decim") and "decim" in inst_data:
                        params["decim"] = inst_data["decim"]
                    if not params.get("n_cycles") and "n_cycles" in inst_data:
                        params["n_cycles"] = inst_data["n_cycles"]
                    if (
                        not params.get("evoked_subtracted")
                        and "evoked_subtracted" in inst_data
                    ):
                        params["evoked_subtracted"] = inst_data["evoked_subtracted"]

                if inst_class:
                    inst = inst_class(
                        name,
                        directory,
                        params.copy(),
                    )
                    subject.add(inst, datatype)

                    # for backwards compatibility,
                    # ensure spectrum objects contain info
                    if datatype == "spectrum" and not params.get("info_set"):
                        inst.set_info(subject)

        experiment.add_subject(subject)

        # ensure that the folder structure exists
        # (to not crash on updates)
        subject.ensure_folders()

    if experiment_updated:
        experiment.save_experiment_settings()

    return experiment



"""Contains plot evoked action handling."""

from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.channels import get_channels_by_type

from meggie.mainwindow.dynamic import Action

from meggie.utilities.dialogs.outputOptionsMain import OutputOptions

from meggie.actions.evoked_plot.controller.evoked import plot_evoked_averages
from meggie.actions.evoked_plot.controller.evoked import plot_evoked_topo


class PlotEvoked(Action):
    """Plots evoked time courses"""

    def run(self, params={}):

        try:
            selected_name = self.data["outputs"]["evoked"][0]
        except IndexError:
            return

        def option_handler(selected_option):
            params = {
                "name": selected_name,
                "output_option": selected_option,
                "channel_groups": self.experiment.channel_groups,
            }
            try:
                self.handler(self.experiment.active_subject, params)
            except Exception as exc:
                exc_messagebox(self.window, exc)

        dialog = OutputOptions(self.window, handler=option_handler)
        dialog.show()

    def handler(self, subject, params):
        """ """
        evoked = subject.evoked.get(params["name"])
        info = evoked.info

        try:
            if params["output_option"] == "channel_averages":
                plot_evoked_averages(evoked, params["channel_groups"])
            else:
                chs = list(get_channels_by_type(info).keys())
                if "eeg" in chs:
                    plot_evoked_topo(subject, evoked, ch_type="eeg")
                if "grad" in chs or "mag" in chs:
                    plot_evoked_topo(subject, evoked, ch_type="meg")
        except Exception as exc:
            exc_messagebox(self.window, exc)



"""Contains a class for logic of the power spectrum creation dialog."""

import logging

from PyQt5 import QtWidgets

import numpy as np

from meggie.utilities.dialogs.powerSpectrumDialogUi import Ui_PowerSpectrumDialog
from meggie.utilities.dialogs.powerSpectrumAddAdvancedDialogMain import (
    PowerSpectrumAddAdvancedDialog,
)

from meggie.utilities.widgets.batchingWidgetMain import BatchingWidget

from meggie.utilities.validators import validate_name
from meggie.utilities.messaging import exc_messagebox
from meggie.utilities.messaging import messagebox


class PowerSpectrumDialog(QtWidgets.QDialog):
    """Contains logic for the power spectrum creation dialog."""

    def __init__(self, experiment, parent, default_name, handler=None):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_PowerSpectrumDialog()
        self.ui.setupUi(self)

        self.parent = parent
        self.experiment = experiment

        self.handler = handler

        self.intervals = []

        raw = self.experiment.active_subject.get_raw()

        tmax = np.floor(raw.times[raw.n_times - 1]) - 0.1
        self.ui.doubleSpinBoxTmin.setValue(0)
        self.ui.doubleSpinBoxTmax.setValue(tmax)
        self.ui.doubleSpinBoxTmin.setMaximum(tmax)
        self.ui.doubleSpinBoxTmax.setMaximum(tmax)

        # set nfft initially to ~2 seconds and overlap to ~1 seconds
        sfreq = raw.info["sfreq"]
        window_in_seconds = 2
        nfft = int(np.power(2, np.ceil(np.log(sfreq * window_in_seconds) / np.log(2))))
        overlap = int(nfft / 2)

        self.ui.spinBoxNfft.setValue(nfft)
        self.ui.spinBoxOverlap.setValue(overlap)

        if raw.info.get("highpass"):
            if self.ui.spinBoxFmin.value() < raw.info["highpass"]:
                self.ui.spinBoxFmin.setValue(int(np.ceil(raw.info["highpass"])))

        if raw.info.get("lowpass"):
            if self.ui.spinBoxFmax.value() > raw.info["lowpass"]:
                self.ui.spinBoxFmax.setValue(int(raw.info["lowpass"]))

        self.batching_widget = BatchingWidget(
            experiment_getter=self._experiment_getter,
            parent=self,
            container=self.ui.groupBoxBatching,
            geometry=self.ui.batchingWidgetPlaceholder.geometry(),
        )
        self.ui.gridLayoutBatching.addWidget(self.batching_widget, 0, 0, 1, 1)

        self.ui.lineEditName.setText(default_name)

    def _experiment_getter(self):
        return self.experiment

    def on_pushButtonAdd_clicked(self, checked=None):
        if checked is None:
            return
        group = str(self.ui.comboBoxAvgGroup.currentText())
        tmin = self.ui.doubleSpinBoxTmin.value()
        tmax = self.ui.doubleSpinBoxTmax.value()
        if tmin >= tmax:
            messagebox(self.parent, "End time must be higher than the starting time")
            return

        self.add_intervals([("fixed", (group, tmin, tmax))])

    def add_intervals(self, intervals):
        """Add intervals to the interval list."""
        for ival_type, interval in intervals:
            self.intervals.append((ival_type, interval))
            if ival_type == "fixed":
                item = QtWidgets.QListWidgetItem(
                    "%s: %s - %s s (fixed)"
                    % (interval[0], round(interval[1], 4), round(interval[2], 4))
                )
            else:
                item_string = str(interval[0]) + ": ["
                if interval[1][0] == "events":
                    item_string += ", ".join([str(elem) for elem in interval[1][1:]])
                elif interval[1][0] == "start":
                    item_string += "start, " + str(interval[1][3])
                elif interval[1][0] == "end":
                    item_string += "end, " + str(interval[1][3])

                item_string += "] – ["

                if interval[2][0] == "events":
                    item_string += ", ".join([str(elem) for elem in interval[2][1:]])
                elif interval[2][0] == "start":
                    item_string += "start, " + str(interval[2][3])

                elif interval[2][0] == "end":
                    item_string += "end, " + str(interval[2][3])

                item_string += "] (dynamic)"

                item = QtWidgets.QListWidgetItem(item_string)
            self.ui.listWidgetIntervals.addItem(item)

    def on_pushButtonClear_clicked(self, checked=None):
        if checked is None:
            return
        self.intervals = []
        self.ui.listWidgetIntervals.clear()

    def on_pushButtonClearRow_clicked(self, checked=None):
        if checked is None:
            return

        current_row = self.ui.listWidgetIntervals.currentRow()
        if current_row == -1:
            return

        self.ui.listWidgetIntervals.takeItem(current_row)
        self.intervals.pop(current_row)

    def on_pushButtonAddAdvanced_clicked(self, checked=None):
        if checked is None:
            return
        dialog = PowerSpectrumAddAdvancedDialog(self)
        dialog.show()

    def accept(self, *args, **kwargs):
        try:
            spectrum_name = validate_name(self.ui.lineEditName.text())
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        intervals = self.intervals
        if not intervals:
            return

        fmin = self.ui.spinBoxFmin.value()
        fmax = self.ui.spinBoxFmax.value()

        subject = self.experiment.active_subject

        params = dict()
        params["fmin"] = fmin
        params["fmax"] = fmax
        params["nfft"] = self.ui.spinBoxNfft.value()
        params["overlap"] = self.ui.spinBoxOverlap.value()
        params["intervals"] = intervals
        params["name"] = spectrum_name

        try:
            self.handler(subject, params)
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        self.parent.initialize_ui()
        self.close()

    def acceptBatch(self, *args):
        try:
            spectrum_name = validate_name(self.ui.lineEditName.text())
        except Exception as exc:
            exc_messagebox(self, exc)
            return

        intervals = self.intervals
        if not intervals:
            return

        fmin = self.ui.spinBoxFmin.value()
        fmax = self.ui.spinBoxFmax.value()

        subject = self.experiment.active_subject

        params = dict()
        params["fmin"] = fmin
        params["fmax"] = fmax
        params["nfft"] = self.ui.spinBoxNfft.value()
        params["overlap"] = self.ui.spinBoxOverlap.value()
        params["intervals"] = intervals
        params["name"] = spectrum_name

        params["run_in_batch"] = True

        selected_subject_names = self.batching_widget.selected_subjects
        for name, subject in self.experiment.subjects.items():
            if name in selected_subject_names:
                try:
                    self.handler(subject, params)
                    subject.release_memory()
                except Exception as exc:
                    self.batching_widget.failed_subjects.append((subject, str(exc)))
                    logging.getLogger("ui_logger").exception("")

        self.batching_widget.cleanup()

        self.parent.initialize_ui()
        self.close()



import tempfile
from meggie.utilities.datasets import get_open_datasets
from meggie.mainwindow.preferences import PreferencesHandler


def test_create_open_datasets():
    tempdir = tempfile.TemporaryDirectory()

    prefs = PreferencesHandler()
    prefs.workspace = tempdir.name

    datasets = get_open_datasets()
    for dataset in datasets.values():
        dataset["constructor"](
            dataset["title"], "", prefs, set_previous_experiment=False
        )

    tempdir.cleanup()



"""Contains a class for logic of the Subjects."""

import os
import logging

import mne

import meggie.utilities.filemanager as filemanager

from meggie.mainwindow.dynamic import find_all_datatype_specs


class Subject:
    """The class for holding subject-specific information
    and subject-specific data.

    Parameters
    ----------
    experiment : meggie.experiment.Experiment
        The experiment to which the subject is created.
    name : str
        Name of the subject.
    raw_fname : str
        Path to the subject data.
    uid : str
        A unique identifier to differentiate between subjects that have
        same name.
    ica_applied : bool
        Whether ICA has been applied (at least once) to this data.
    rereferenced : bool
        Whether the data has been rereferenced (at least once).
    """

    def __init__(
        self, experiment, name, raw_fname, uid, ica_applied=False, rereferenced=False
    ):
        self.name = name
        self.raw_fname = raw_fname

        self.uid = uid

        self._raw = None

        self.ica_applied = ica_applied
        self.rereferenced = rereferenced

        self.path = os.path.join(experiment.path, name)

        datatype_specs = find_all_datatype_specs()
        for source, package, datatype_spec in datatype_specs.values():
            datatype = datatype_spec["id"]
            dir_ = datatype_spec["dir"]
            setattr(self, datatype, dict())
            setattr(self, datatype + "_directory", os.path.join(self.path, dir_))

    def add(self, dataobject, datatype):
        """Adds a dataobject of type datatype to the subject.

        Parameters
        ----------
        dataobject : instance of a datatype
            A data object.
        datatype : str
            Name of the datatype.
        """
        container = getattr(self, datatype)
        name = dataobject.name
        container[name] = dataobject

    def remove(self, name, datatype):
        """Removes a dataobject by name from the subject.

        Parameters
        ----------
        name : str
            Name of the data object.
        datatype : str
            Name of the datatype.

        """
        container = getattr(self, datatype)
        dataobject = container.pop(name, None)
        try:
            dataobject.delete_content()
        except Exception:
            logging.getLogger("ui_logger").exception("")
            raise IOError("Could not delete " + str(datatype) + " from folders")

    @property
    def raw_path(self):
        """Returns the raw path."""
        path = os.path.join(self.path, self.raw_fname)
        return path

    def get_raw(self, preload=True, verbose="warning"):
        """Gets the raw object for the subject.

        Reads from the file system if not in the memory already.

        Parameters
        ----------
        preload : bool
            Whether to read the data or only the metadata.
        verbose : str
            Verbose level of read_raw.

        Returns
        -------
        mne.io.Raw
            The raw object.
        """
        if self._raw is not None:
            if preload:
                self._raw.load_data()
            return self._raw
        else:
            try:
                raw = filemanager.open_raw(
                    self.raw_path, preload=preload, verbose=verbose
                )
            except OSError:
                raise IOError("Could not find the raw file.")
            self._raw = raw
            return raw

    def save(self):
        """Saves the data to the existing path."""
        if self.has_raw:
            try:
                filemanager.save_raw(self._raw, self.raw_path)
            except Exception:
                raise Exception(
                    "Could not save the raw file. Please ensure "
                    "that the entire experiment folder has "
                    "write permissions."
                )

    def release_memory(self):
        """Releases data from the memory."""
        if self._raw is not None:
            self._raw = None

    @property
    def has_eeg(self):
        """Checks if the raw has eeg data present"""
        raw = self.get_raw(preload=False)
        channels = mne.pick_types(raw.info, eeg=True, meg=False)
        if len(channels) == 0:
            return False
        return True

    @property
    def has_raw(self):
        """Checks if is a 'placeholder' subject."""
        if self.raw_fname:
            return True
        else:
            return False

    @property
    def sss_applied(self):
        """Checks if sss applied."""

        try:
            raw = self.get_raw()
            for item in raw.info["proc_history"]:
                if "maxfilter" in item.get("creator", []):
                    return True
        except Exception:
            return False

        return False

    def ensure_folders(self):
        """When called, checks that the subject folder with all datatype folders
        exist and if not, creates them.
        """
        paths = []
        datatype_specs = find_all_datatype_specs()

        for source, package, datatype_spec in datatype_specs.values():
            datatype = datatype_spec["id"]
            path = getattr(self, datatype + "_directory")
            paths.append(path)

        try:
            filemanager.ensure_folders([self.path] + paths)
        except OSError:
            raise OSError(
                "Couldn't create all the necessary folders. "
                "Please ensure that the experiment folder "
                "has write permissions everywhere."
            )

    def save_montage(self, montage):

        ch_names = montage.ch_names
        with open(os.path.join(self.path, "montage_channels.txt"), "w") as f:
            for ch_name in ch_names:
                f.write(ch_name + "\n")

        montage.save(os.path.join(self.path, "montage.fif"), overwrite=True)

    def read_montage(self):
        montage_path = os.path.join(self.path, "montage.fif")
        channels_path = os.path.join(self.path, "montage_channels.txt")
        if os.path.exists(montage_path) and os.path.exists(channels_path):
            ch_names = []
            with open(channels_path, "r") as f:
                for line in f:
                    ch_names.append(line.strip())

            montage = mne.channels.read_dig_fif(montage_path)

            mapping = {}
            for idx, name in enumerate(montage.ch_names):
                mapping[name] = ch_names[idx]

            montage.rename_channels(mapping)

            return montage



"""Contains controlling logic for the tfr implementation"""

import numpy as np
import matplotlib.pyplot as plt
import mne

from mne._fiff.pick import channel_type

from meggie.utilities.plotting import color_cycle
from meggie.utilities.plotting import create_channel_average_plot
from meggie.utilities.plotting import set_figure_title
from meggie.utilities.channels import average_to_channel_groups
from meggie.utilities.channels import iterate_topography
from meggie.utilities.channels import filter_info
from meggie.utilities.channels import ensure_montage
from meggie.utilities.units import get_power_unit


def _compute_tse(meggie_tfr, fmin, fmax):
    tfrs = meggie_tfr.content

    fmin_idx = np.where(meggie_tfr.freqs >= fmin)[0][0]
    fmax_idx = np.where(meggie_tfr.freqs <= fmax)[0][-1]

    if fmax_idx <= fmin_idx:
        raise Exception("Something wrong with the frequencies")

    tses = {}
    for key, tfr in tfrs.items():
        tse = np.mean(tfr.data[:, fmin_idx : fmax_idx + 1, :], axis=1)
        tses[key] = tse

    return tses


def _crop_and_correct_to_baseline(tse, blmode, blstart, blend, tmin, tmax, times):
    tmin_idx = np.where(times >= tmin)[0][0]
    tmax_idx = np.where(times <= tmax)[0][-1]

    if blmode:
        if blstart < tmin:
            raise Exception("Baseline start should not be earlier than crop start.")

        if blend > tmax:
            raise Exception("Baseline end should not be later than crop end.")

        # correct to baseline
        tse = mne.baseline.rescale(tse, times, baseline=(blstart, blend), mode=blmode)

        # crop
        tse = tse[:, tmin_idx : tmax_idx + 1]

    return times[tmin_idx : tmax_idx + 1], tse


def plot_tse_topo(
    subject, tfr_name, blmode, blstart, blend, tmin, tmax, fmin, fmax, ch_type
):
    """Plots a tse topography."""
    meggie_tfr = subject.tfr.get(tfr_name)

    tses = _compute_tse(meggie_tfr, fmin, fmax)

    info = meggie_tfr.info
    if ch_type == "meg":
        picked_channels = [
            ch_name
            for ch_idx, ch_name in enumerate(info["ch_names"])
            if ch_idx in mne.pick_types(info, meg=True, eeg=False)
        ]
    else:
        picked_channels = [
            ch_name
            for ch_idx, ch_name in enumerate(info["ch_names"])
            if ch_idx in mne.pick_types(info, meg=False, eeg=True)
        ]

    info = filter_info(info, picked_channels)

    ensure_montage(subject, info, ch_type)

    ch_names = meggie_tfr.ch_names
    colors = color_cycle(len(tses))

    def individual_plot(ax, info_idx, names_idx):
        """ """
        ch_name = ch_names[names_idx]

        title = " ".join([tfr_name, ch_name])
        set_figure_title(ax.figure, title.replace(" ", "_"))
        ax.figure.suptitle(title)
        ax.set_title("")

        for color_idx, (key, tse) in enumerate(sorted(tses.items())):
            times, tse = _crop_and_correct_to_baseline(
                tse, blmode, blstart, blend, tmin, tmax, meggie_tfr.times
            )
            ax.plot(times, tse[names_idx], color=colors[color_idx], label=key)
            ax.axhline(0, color="black")
            ax.axvline(0, color="black")

        ax.legend()

        ax.set_xlabel("Time (s)")

        if blmode and blmode != "mean":
            ax.set_ylabel("Power")
        else:
            ax.set_ylabel(
                "Power ({})".format(
                    get_power_unit(channel_type(info, info_idx), log=False)
                )
            )

        plt.show()

    fig = plt.figure()
    for ax, info_idx, names_idx in iterate_topography(
        fig, info, ch_names, individual_plot
    ):

        handles = []
        for color_idx, (key, tse) in enumerate(sorted(tses.items())):
            times, tse = _crop_and_correct_to_baseline(
                tse, blmode, blstart, blend, tmin, tmax, meggie_tfr.times
            )
            handles.append(
                ax.plot(
                    times,
                    tse[names_idx],
                    linewidth=0.2,
                    color=colors[color_idx],
                    label=key,
                )[0]
            )

    fig.legend(handles=handles)
    title = "{0}_{1}".format(tfr_name, ch_type)
    set_figure_title(fig, title)
    plt.show()


def plot_tse_averages(
    subject, tfr_name, blmode, blstart, blend, tmin, tmax, fmin, fmax, channel_groups
):
    """Plots tse averages."""
    meggie_tfr = subject.tfr.get(tfr_name)

    tses = _compute_tse(meggie_tfr, fmin, fmax)

    ch_names = meggie_tfr.ch_names
    info = meggie_tfr.info
    colors = color_cycle(len(tses))
    conditions = meggie_tfr.content.keys()

    averages = {}
    for key, tse in sorted(tses.items()):

        # note: baseline is corrected before channel average
        times, ch_data = _crop_and_correct_to_baseline(
            tse, blmode, blstart, blend, tmin, tmax, meggie_tfr.times
        )

        data_labels, averaged_data = average_to_channel_groups(
            ch_data, info, ch_names, channel_groups
        )

        for label_idx, label in enumerate(data_labels):
            if label not in averages:
                averages[label] = []
            averages[label].append((key, times, averaged_data[label_idx]))

    ch_types = sorted(set([label[0] for label in averages.keys()]))
    for ch_type in ch_types:

        ch_groups = sorted(
            [label[1] for label in averages.keys() if label[0] == ch_type]
        )

        def plot_fun(ax_idx, ax):
            ch_group = ch_groups[ax_idx]
            ax.set_title(ch_group)
            ax.set_xlabel("Time (s)")

            if blmode and blmode != "mean":
                ax.set_ylabel("Power")
            else:
                ax.set_ylabel("Power ({})".format(get_power_unit(ch_type, log=False)))

            for color_idx, (key, times, curve) in enumerate(
                averages[(ch_type, ch_group)]
            ):
                ax.plot(times, curve, color=colors[color_idx], label=key)

            ax.axhline(0, color="black")
            ax.axvline(0, color="black")

        title = " ".join([tfr_name, ch_type])
        legend = list(zip(conditions, colors))
        create_channel_average_plot(len(ch_groups), plot_fun, title, legend)

    plt.show()



from meggie.utilities.testing import BaseTestAction
from meggie.actions.tfr_delete_from_all import DeleteTFRFromAll


class TestTFRDeleteFromAll(BaseTestAction):
    def test_tfr_delete_from_all(self):

        data = {"outputs": {"tfr": ["TFR"]}}

        self.run_action(
            action_name="tfr_delete_from_all",
            handler=DeleteTFRFromAll,
            data=data,
        )



from meggie.utilities.testing import BaseTestAction
from meggie.actions.raw_measurement_info import Info


class TestRawMeasurementInfo(BaseTestAction):
    def test_raw_measurement_info(self):

        content = self.run_action(
            action_name="raw_measurement_info",
            handler=Info,
        )
        assert content



"""Contains controlling logic for the epochs creation."""

import copy
import logging

import numpy as np
import mne

from meggie.utilities.units import get_scaling
from meggie.utilities.events import find_stim_channel
from meggie.utilities.events import find_events

from meggie.datatypes.epochs.epochs import Epochs


def create_epochs_from_events(subject, params):
    """Creates epochs based on events."""
    raw = subject.get_raw()

    params = copy.deepcopy(params)
    event_params = params["events"]
    reject_params = params["reject"]

    # convert reject params from human readable units to standard units
    for key in reject_params:
        reject_params[key] /= get_scaling(key)

    # remove params that don't match with the channel types present in raw
    if mne.pick_types(raw.info, meg="grad", eeg=False).size == 0:
        reject_params.pop("grad", None)
    if mne.pick_types(raw.info, meg="mag", eeg=False).size == 0:
        reject_params.pop("mag", None)
    if mne.pick_types(raw.info, meg=False, eeg=True).size == 0:
        reject_params.pop("eeg", None)

    events = []
    category = {}

    stim_channel = find_stim_channel(raw)

    # event_id should not matter after epochs are created,
    # so we just add placeholders as 1, 2, 3...
    if len(event_params) > 0:
        for idx, item in enumerate(event_params):
            event_id = item["event_id"]
            mask = item["mask"]
            category_id = "id_" + str(event_id) + "_mask_" + str(mask)

            new_events = find_events(raw, stim_channel, mask, event_id)

            if len(new_events) == 0:
                logging.warning("No events found with setting " + str(category_id))
                continue

            category[category_id] = idx + 1
            new_events[:, 2] = idx + 1

            events.extend(
                [
                    [
                        event[0] + int(round(raw.info["sfreq"] * params["delay"])),
                        event[1],
                        event[2],
                    ]
                    for event in new_events
                ]
            )

    if len(events) == 0:
        raise ValueError(
            "No matching events found. Please check rejection limits and other parameters."
        )

    # prepare parameters for pick_types
    if params["mag"] and params["grad"]:
        meg = True
    elif params["mag"]:
        meg = "mag"
    elif params["grad"]:
        meg = "grad"
    else:
        meg = False

    eeg = params["eeg"]

    # find all proper picks, dont exclude bads
    picks = mne.pick_types(raw.info, meg=meg, eeg=eeg, exclude=[])

    if len(picks) == 0:
        raise ValueError("You should select at least one channel type")

    mne_epochs = mne.Epochs(
        raw,
        np.array(events),
        category,
        params["tmin"],
        params["tmax"],
        baseline=(params["bstart"], params["bend"]),
        picks=picks,
        reject=reject_params,
    )

    if len(mne_epochs.get_data()) == 0:
        raise ValueError(
            "Could not find any data. Perhaps the "
            + "rejection thresholds are too strict..."
        )

    n_dropped = len(events) - len(mne_epochs.get_data())

    if n_dropped > 0:
        logging.getLogger("ui_logger").info(str(n_dropped) + " epochs dropped.")

    epochs_directory = subject.epochs_directory
    epochs = Epochs(
        params["collection_name"], epochs_directory, params, content=mne_epochs
    )

    epochs.save_content()
    subject.add(epochs, "epochs")



"""Contains create spectrum action handling."""

from meggie.utilities.names import next_available_name

from meggie.mainwindow.dynamic import Action
from meggie.mainwindow.dynamic import subject_action

from meggie.utilities.dialogs.powerSpectrumDialogMain import PowerSpectrumDialog
from meggie.actions.spectrum_create.controller.spectrum import create_power_spectrum


class CreateSpectrum(Action):
    """Shows a dialog to get parameters for spectrum creation and
    then allows creating spectrums"""

    def run(self, params={}):
        default_name = next_available_name(
            self.experiment.active_subject.spectrum.keys(), "Spectrum"
        )

        dialog = PowerSpectrumDialog(
            self.experiment, self.window, default_name, self.handler
        )
        dialog.show()

    @subject_action
    def handler(self, subject, params):
        """ """
        spectrum_name = params["name"]
        intervals = params["intervals"]
        fmin = params["fmin"]
        fmax = params["fmax"]
        nfft = params["nfft"]
        overlap = params["overlap"]

        create_power_spectrum(
            subject,
            spectrum_name,
            intervals,
            fmin,
            fmax,
            nfft,
            overlap,
            do_meanwhile=self.window.update_ui,
        )
        self.experiment.save_experiment_settings()



"""Contains a logic for setting active plugins."""

from PyQt5 import QtWidgets, QtCore

from meggie.mainwindow.dynamic import find_all_plugins
from meggie.mainwindow.dynamic import find_all_package_specs

from meggie.mainwindow.dialogs.activePluginsDialogUi import Ui_activePluginsDialog


class ActivePluginsDialog(QtWidgets.QDialog):
    """Contains logic for active plugins dialog."""

    def __init__(self, active_plugins, parent, handler):
        QtWidgets.QDialog.__init__(self, parent)
        self.ui = Ui_activePluginsDialog()
        self.ui.setupUi(self)

        self.handler = handler

        plugins = find_all_plugins()
        package_specs = find_all_package_specs()

        plugin_info = []
        for source, package_spec in package_specs.items():
            if source not in plugins:
                continue

            id_ = package_spec["id"]
            name = package_spec["name"]
            author = package_spec["author"]

            plugin_info.append((id_, name, author))
        self.plugin_info = plugin_info

        for id_, name, author in plugin_info:
            text = name + " - " + author
            item = QtWidgets.QListWidgetItem(text)
            self.ui.listWidgetPlugins.addItem(item)
            if id_ in active_plugins:
                item.setSelected(True)

        # Create the hyperlink label
        self.link_label = QtWidgets.QLabel(self)
        self.link_label.setTextFormat(QtCore.Qt.RichText)
        self.link_label.setTextInteractionFlags(QtCore.Qt.TextBrowserInteraction)
        self.link_label.setOpenExternalLinks(True)

        # Define URL as a variable for reuse
        plugin_url = "https://cibr-jyu.github.io/meggie/user-guide/plugins"

        self.link_label.setText(
            f'Find and install plugins: <a href="{plugin_url}">Documentation</a>'
        )

        # Add tooltip with full URL
        self.link_label.setToolTip(plugin_url)

        # Insert the link label below the plugin list
        self.ui.gridLayout.addWidget(self.link_label, 1, 0)

        # Move buttons to next row
        self.ui.gridLayout.removeItem(self.ui.horizontalLayout)
        self.ui.gridLayout.addLayout(self.ui.horizontalLayout, 2, 0)

    def accept(self):
        idxs = [elem.row() for elem in self.ui.listWidgetPlugins.selectedIndexes()]
        active_plugins = [self.plugin_info[idx][0] for idx in idxs]
        self.handler(active_plugins)
        self.close()



from meggie.utilities.names import next_available_name


def test_next_available_name():
    names = ["kissa", "koira", "kissa_1", "kissa_2", "kissa_11", "kissa_kala"]
    assert next_available_name(names, "kissa") == "kissa_12"
    assert next_available_name(names, "koira") == "koira_1"
    assert next_available_name(names, "kettu") == "kettu"
    assert next_available_name(names, "kissa_kala") == "kissa_kala_1"

    names = ["EO", "EOEC", "group_EOEC"]
    assert next_available_name(names, "group_EO") == "group_EO"



from PyQt5 import QtCore

from meggie.utilities.testing import BaseTestAction
from meggie.actions.tfr_group_average import GroupAverage
from meggie.utilities.dialogs.groupSelectionDialogMain import (
    GroupSelectionDialog,
)


class TestTFRGroupAverage(BaseTestAction):
    def test_tfr_group_average(self):

        data = {"outputs": {"tfr": ["TFR"]}}

        self.run_action(
            action_name="tfr_group_average",
            handler=GroupAverage,
            data=data,
        )
        dialog = self.find_dialog(GroupSelectionDialog)

        dialog.ui.checkBoxGroup_0.setCheckState(QtCore.Qt.CheckState.Checked)
        dialog.ui.spinBoxGroup_0.setValue(1)
        dialog.ui.checkBoxGroup_1.setCheckState(QtCore.Qt.CheckState.Checked)
        dialog.ui.spinBoxGroup_1.setValue(1)

        dialog.accept()



"""Useful functions for formatting outputs."""


def format_floats(data):
    """Returns formatted list of floats from unformatted list of floats

    Parameters
    ----------
    data : list
        The list of floats to be formatted

    Returns
    -------
    list
        Formatted floats (now string).
    """
    return ["{0:.3f}".format(elem) for elem in data]


def format_float(value):
    """Formats a float.

    Parameters
    ----------
    value : float
        Float to be formatted

    Returns
    -------
    str
        Formatted float.

    """
    return format_floats([value])[0]



import tempfile
import os

import mne

from meggie.datatypes.spectrum.spectrum import Spectrum

from meggie.utilities.filemanager import ensure_folders


def test_spectrum():
    with tempfile.TemporaryDirectory() as dirpath:

        sample_folder = mne.datasets.sample.data_path()
        sample_fname = os.path.join(
            sample_folder, "MEG", "sample", "sample_audvis_raw.fif"
        )

        raw = mne.io.read_raw_fif(sample_fname, preload=True)
        mne_spectrum = raw.compute_psd(fmin=1, fmax=40, tmin=1, tmax=10)
        psds = mne_spectrum.get_data()
        freqs = mne_spectrum.freqs

        name = "TestSpectrum"
        cond_name = "1"
        spectrum_dir = os.path.join(dirpath, "spectrums")

        # one meggie-Spectrum can hold many spectrums, thus content is dict-like
        content = {cond_name: psds}
        params = {"conditions": [cond_name]}

        # Create meggie-Spectrum object with spectrum array stored within
        # and save it to spectrum directory
        spectrum = Spectrum(
            name, spectrum_dir, params, content=content, freqs=freqs, info=raw.info
        )
        ensure_folders([spectrum_dir])
        spectrum.save_content()

        # Creating meggie-Spectrum object with same name and folder should allow
        # accessing the saved content
        loaded_spectrum = Spectrum(name, spectrum_dir, params)

        assert list(loaded_spectrum.content.keys())[0] == cond_name



from meggie.utilities.testing import BaseTestAction
from meggie.actions.epochs_delete import DeleteEpochs


class TestEpochsDelete(BaseTestAction):
    def test_epochs_delete(self):

        data = {"outputs": {"epochs": ["Epochs"]}}

        self.run_action(
            action_name="epochs_delete",
            handler=DeleteEpochs,
            data=data,
        )



"""Contains implementation for spectrum info"""

from meggie.mainwindow.dynamic import InfoAction

from meggie.utilities.formats import format_float


class Info(InfoAction):
    """Fills up spectrum info box"""

    def run(self, params={}):
        try:
            selected_name = self.data["outputs"]["spectrum"][0]

            spectrum = self.experiment.active_subject.spectrum[selected_name]
            params = spectrum.params

            message = ""

            message += "Name: " + spectrum.name + "\n\n"

            if "fmin" in params and "fmax" in params:
                message += "Frequencies: {0}Hz - {1}Hz\n".format(
                    format_float(params["fmin"]), format_float(params["fmax"])
                )

            if "nfft" in params:
                message += "Window length (samples): {0}\n".format(params["nfft"])

            if "overlap" in params:
                message += "Overlap (samples): {0}\n".format(params["overlap"])

            if "intervals" in params:
                message += "\nIntervals: \n"
                for key, ivals in params["intervals"].items():
                    message += "Condition " + str(key) + ": "
                    message += ", ".join(
                        [
                            "({0}s - {1}s)".format(
                                format_float(ival[0]), format_float(ival[1])
                            )
                            for ival in ivals
                        ]
                    )
                    message += "\n"

            if "groups" in params:
                for key, names in params["groups"].items():
                    message += "\nGroup " + str(key) + ": \n"
                    for name in names:
                        message += name + "\n"

        except Exception:
            message = ""

        return message



from meggie.utilities.testing import BaseTestAction
from meggie.actions.evoked_delete import DeleteEvoked


class TestEvokedDelete(BaseTestAction):
    def test_evoked_delete(self):

        data = {"outputs": {"evoked": ["Evoked"]}}

        self.run_action(
            action_name="evoked_delete",
            handler=DeleteEvoked,
            data=data,
        )
